#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/InputArtifact.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  13.08.2022
# Last Modified By:    jvs

from pathlib import Path
from pydantic import root_validator
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.job.InputArtifactHttpSource import InputArtifactHttpSource
from ampel.model.job.utils import transform_expressions

class InputArtifact(AmpelBaseModel):

    name: str
    path: Path
    http: InputArtifactHttpSource

    def value(self) -> None | str:
        try:
            return self.path.read_text()
        except FileNotFoundError:
            return None

    @staticmethod
    def _check_expression(expression: str):
        """For compatibility with Argo, forbid use of {{ item }} in artifact specs"""
        if expression == "item":
            raise ValueError("artifact fields cannot reference {{ item }}")
        return expression

    @root_validator
    def check_for_item(cls, values):
        transform_expressions(values, cls._check_expression)
        return values



import asyncio
from contextlib import asynccontextmanager

import pytest

from ampel.core.DefaultProcessController import DefaultProcessController
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry


@asynccontextmanager
async def run(controller):
    task = asyncio.create_task(controller.run())
    try:
        yield task
    finally:
        controller.stop()
        await task


@pytest.mark.asyncio
async def test_process_gauge(dev_context, ampel_logger):

    process_count = lambda: AmpelMetricsRegistry.registry().get_sample_value(
        "ampel_processes", {"tier": "2", "process": "sleepy"}
    )

    c = DefaultProcessController(
        config=dev_context.config,
        processes=[
            {
                "name": "sleepy",
                "schedule": "every(30).seconds",
                "tier": 2,
                "isolate": True,
                "version": 0,
                "processor": {"unit": "Sleepy"},
            }
        ],
    )
    async with run(c) as task:
        try:
            await asyncio.wait_for(asyncio.shield(task), 1)
        except asyncio.exceptions.TimeoutError:
            ...
        assert process_count() == 1
    assert process_count() == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/DBCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.03.2021
# Last Modified Date:  20.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from collections.abc import Sequence
from argparse import ArgumentParser # type: ignore[import]
from ampel.core.AmpelDB import AmpelDB
from ampel.core.AmpelContext import AmpelContext
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.main import AmpelArgumentParser
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag


# Help parameter descriptions
hlp = {
	'import': 'Import saved ampel databases',
	'save': 'Save a databases into file',
	'delete': 'Deletes all database starting with the provided prefix (ex: AmpelTest)',
	'prefix': 'Prefix of the collections to delete (ex: AmpelTest)',
	'config': 'Path to an ampel config file (yaml/json)',
	'secrets': 'Path to a YAML secrets store in sops format',
	'log-profile': 'One of: default, compact, headerless, verbose, debug',
	'debug': 'debug',
	'force': 'Delete potententially existing view before view creation',
	'view': 'Create or discard collection views',
	'index': 'Create or recreate collection indexes',
}

class DBCommand(AbsCoreCommand):


	@staticmethod
	def get_sub_ops() -> list[str]:
		return ['import', 'export', 'delete', 'view', 'index']


	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = self.get_sub_ops()
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'db', sub_ops, hlp, description = 'Import, export or delete ampel databases. Create or remove views.'
			)

		builder = ArgParserBuilder('db')
		builder.add_parsers(sub_ops, hlp)
		builder.notation_add_note_references()
		builder.notation_add_example_references()

		builder.req('config')
		builder.req('in', 'import')
		builder.req('out', 'export')

		builder.add_group('view', 'View arguments', sub_ops='view')
		builder.xargs(
			group='required', sub_ops='view', xargs = [
				{'name': 'create', 'action': 'store_true'},
				{'name': 'discard', 'action': 'store_true'}
			]
		)
		builder.xargs(
			group='required', sub_ops='view', xargs = [
				{'name': 'channel'},
				{'name': 'channels-or', 'nargs': '+'},
				{'name': 'channels-and', 'nargs': '+'}
			]
		)

		# Optional
		builder.opt('secrets')
		builder.opt('one-db', default=False, action='store_true')
		builder.opt('debug', action='store_true')
		builder.opt('force', 'view|index', action='store_true')
		builder.opt('col', 'index', action='append')

		builder.example('import', '-in /path/to/file')
		builder.example('export', '-out /path/to/file')
		builder.example('delete', '-mongo.prefix AmpelTest')
		builder.example('view', '-create -channel CHAN1')
		builder.example('view', '-create -channels-or CHAN1 CHAN2')
		builder.example('view', '-discard -channel CHAN1')

		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		if sub_op == 'delete':  # cosmetic mainly
			AmpelDB.create_collection = (lambda x: None) # type: ignore

		ctx: AmpelContext = self.get_context(args, unknown_args, one_db=args['one_db'], require_existing_db=sub_op != 'index')
		db = ctx.db

		logger = AmpelLogger.from_profile(
			ctx, 'console_debug' if args['debug'] else 'console_info',
			base_flag = LogFlag.MANUAL_RUN
		)

		if sub_op == 'delete':

			logger.info(f'Deleting databases with prefix {ctx.db.prefix}')
			ctx.db.drop_all_databases()
			logger.info('Done')

		elif sub_op == 'index':

			for col_name, col_model in ctx.db.col_config.items():
				if args['col'] and not col_name in args['col']:
					continue
				col = ctx.db.get_collection(col_name)
				ctx.db.set_col_index(col, col_model, logger, force_overwrite=args['force'])

		elif sub_op == 'view':

			if 'create' not in args and 'discard' not in args:
				logger.error('Either provide "create" or "discard" in combination with view command')
				return

			try:
				if x := args.get('channel'):
					logger.info(f'{"Creating" if args["create"] else "Removing"} view for channel {x}')
					db.create_one_view(x, logger, args['force']) if args['create'] else db.delete_one_view(x, logger)
				elif x := args.get('channels_or'):
					logger.info(f'{"Creating" if args["create"] else "Removing"} view for channels {x}')
					db.create_or_view(x, logger, args['force']) if args['create'] else db.delete_or_view(x, logger)
				elif x := args.get('channels_and'):
					logger.info(f'{"Creating" if args["create"] else "Removing"} view for channels {x}')
					db.create_and_view(x, logger, args['force']) if args['create'] else db.delete_and_view(x, logger)
				else:
					logger.error('Channel(s) required\n')
					return
			except Exception as e:
				print(e)
				if 'already exists' in str(e):
					#logger.info('View already exist')
					logger.info(str(e))
				else:
					raise e

		elif sub_op == 'export':
			raise NotImplementedError()

		elif sub_op == 'import':
			raise NotImplementedError()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/export.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.08.2022
# Last Modified Date:  16.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys
from json import dumps
from bson import ObjectId, encode
from io import BufferedWriter, TextIOWrapper
from itertools import islice
from datetime import datetime
from typing import TextIO, BinaryIO
from collections.abc import Generator

from ampel.protocol.LoggerProtocol import LoggerProtocol
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.util.serialize import walk_and_encode
from ampel.util.pretty import prettyjson
from ampel.util.getch import getch as fgetch

dsi = dict.__setitem__
ufts = datetime.utcfromtimestamp


def txt_export(
	fd: TextIOWrapper | TextIO,
	gen: Generator[AmpelBuffer, None, None],
	id_mapper: None | AbsIdMapper = None,
	chunk_size: int = 200,
	human_times: bool = True,
	pretty: bool = False,
	getch: bool = False,
	close_fd: bool = True,
	logger: None | LoggerProtocol = None
) -> None:

	first = True
	func = prettyjson if pretty else dumps

	fd.write('[\n')

	try:

		while (data := list(islice(gen, chunk_size))):

			walk_and_encode(data)

			if first and len(data) > 0:

				data = iter(data) # type: ignore
				el = next(data) # type: ignore
				if id_mapper:
					el['id'] = id_mapper.to_ext_id(el['id'])
				if human_times:
					convert_timestamps(el)
				if logger:
					logger.info(f"Writing content (id: {el['id']})") # type: ignore
				fd.write(func(el)) # type: ignore
				first = False

			if getch and fgetch():
				fd.write('\n]\n')
				if logger:
					logger.info("Abording")
				return None

			for el in data:

				fd.write(",\n")
				if id_mapper:
					el['id'] = id_mapper.to_ext_id(el['id'])
				if human_times:
					convert_timestamps(el)
				if logger:
					logger.info(f"Writing content (id: {el['id']})") # type: ignore
				fd.write(func(el)) # type: ignore

				if getch and fgetch():
					fd.write('\n]\n')
					if logger:
						logger.info("Abording")
					return None

		fd.write('\n]\n')

	finally:

		if close_fd:
			fd.flush()
			fd.close()


def bin_export(
	fd: BufferedWriter | BinaryIO,
	gen: Generator[AmpelBuffer, None, None],
	id_mapper: None | AbsIdMapper = None,
	close_fd: bool = True,
	logger: None | LoggerProtocol = None
) -> None:

	for el in gen:
		if logger:
			logger.info(f"Writing content (id: {el['id']})") # type: ignore[str-bytes-safe]
		if id_mapper:
			el['id'] = id_mapper.to_ext_id(el['id'])
		fd.write(encode(el))

	if close_fd:
		if logger:
			logger.info("Closing file descriptor")
		fd.flush()
		fd.close()


def get_fd(
	binary: bool,
	fd: None | str | BufferedWriter | TextIO | TextIOWrapper | BinaryIO,
	close_fd: bool
) -> tuple[BufferedWriter | TextIO | TextIOWrapper | BinaryIO, bool]:

	if binary:
		if isinstance(fd, str):
			return open(fd, "wb"), True
		return fd, close_fd # type: ignore
	else:
		if fd is None:
			return sys.stdout, False
		if isinstance(fd, str):
			return open(fd, 'w'), close_fd
		return fd, close_fd # type: ignore



def convert_timestamps(ab: AmpelBuffer) -> None:

	try:

		if stock_doc := ab.get('stock'):
			for k in (ts := stock_doc['ts']):
				dsi( # type: ignore[misc]
					ts, k, {
						'tied': ufts(ts[k]['tied']).isoformat(),
						'upd': ufts(ts[k]['upd']).isoformat()
					}
				)
			for je in stock_doc['journal']:
				dsi(je, 'ts', ufts(je['ts']).isoformat()) # type: ignore[arg-type]

		if t1s := ab.get('t1'):
			for t1doc in t1s:
				if isinstance(t1doc['_id'], str) and "oid:" in t1doc['_id']: # type: ignore[typeddict-item]
					gt = ObjectId(t1doc['_id'][4:]).generation_time # type: ignore[typeddict-item]
				else:
					gt = t1doc['_id'].generation_time # type: ignore[typeddict-item, union-attr]
				dsi(t1doc, 'added', gt.isoformat()) # type: ignore[arg-type, typeddict-item]
				for t1meta in t1doc['meta']:
					dsi(t1meta, 'ts', ufts(t1meta['ts']).isoformat()) # type: ignore[arg-type]

		if t2s := ab.get('t2'):
			for t2doc in t2s:
				for t2meta in t2doc['meta']:
					dsi(t2meta, 'ts', ufts(t2meta['ts']).isoformat()) # type: ignore[arg-type]

	except Exception:
		import traceback, sys
		traceback.print_exc()
		print(ab)
		sys.exit("Exception occured")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/demo/DemoPlainT3Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.12.2021
# Last Modified Date:  17.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.struct.T3Store import T3Store
from ampel.abstract.AbsT3PlainUnit import AbsT3PlainUnit


class DemoPlainT3Unit(AbsT3PlainUnit):

	a_parameter: int = 9000
	my_t3_doc_tag: str = "A_TAG"


	def post_init(self) -> None:
		self.logger.info("post_init was called")


	def process(self, t3s: T3Store) -> UBson | UnitResult:

		self.logger.info("Running DemoPlainT3Unit")

		if not t3s.units:
			self.logger.info(f"T3 store contains t3 views for units: {t3s.units}")
		else:
			self.logger.info("T3 store contains no t3 views")

		return UnitResult(
			body = {'a_parameter': self.a_parameter},
			tag = self.my_t3_doc_tag
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/AbsDictConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.03.2020
# Last Modified Date:  03.03.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.config.collector.ConfigCollector import ConfigCollector


class AbsDictConfigCollector(ConfigCollector, AmpelABC, abstract=True):

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		# value: tuple[dist_name, version, register_file]
		self._origin: dict[str, tuple[str, None | str | float | int, str]] = {}


	def check_duplicates(self,
		key: str,
		dist_name: str,
		version: str | float | int,
		register_file: str,
		section_detail: None | str = None
	) -> bool:
		if self._origin.get(key):
			self.report_duplicated_entry(
				conf_key = key,
				section_detail = section_detail or f'{self.tier} {self.conf_section}',
				new_dist = dist_name,
				new_file = register_file,
				prev_dist = self._origin[key][0],
				prev_file = self._origin[key][2]
			)
			return True

		self._origin[key] = (dist_name, version, register_file)
		return False


	@abstractmethod
	def add(self,
		arg: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None | int:
		...



import copy

from ampel.config.AmpelConfig import AmpelConfig
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.test.dummy import DummyStateT2Unit


def test_load_old_configids(mock_context: DevAmpelContext, ampel_logger):
    """
    Configuration hashes stored via hash_ingest_directive are loaded from the
    database when a new DevAmpelContext is instantiated
    """
    unit_config = {"foo": 37}
    mock_context.register_unit(DummyStateT2Unit)
    pre_register_context = DevAmpelContext(
        config=AmpelConfig(copy.deepcopy(mock_context.config.get())),
        db=mock_context.db,
        loader=mock_context.loader,
    )
    mock_context.hash_ingest_directive(
        dict(
            channel="TEST",
            ingest=dict(
                combine=[
                    dict(
                        unit="T1SimpleCombiner",
                        state_t2=[dict(unit="DummyStateT2Unit", config=unit_config)],
                    )
                ]
            ),
        ),
        ampel_logger,
    )
    post_register_context = DevAmpelContext(
        config=AmpelConfig(copy.deepcopy(pre_register_context.config.get())),
        db=mock_context.db,
        loader=mock_context.loader,
    )
    assert (
        isinstance(
            pre_register_confid := pre_register_context.config.get("confid", dict), dict
        )
        and len(pre_register_confid) == 0
    ), "config not present before registration"

    assert (
        isinstance(confid := mock_context.config.get("confid", dict), dict)
        and len(confid) == 1
    ), "config present after registration"
    assert next(iter(confid.values())) == unit_config
    assert (
        post_register_context.config.get("confid", dict) == confid
    ), "unregistered configs loaded from database"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/metrics/AmpelDBCollector.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  11.02.2021
# Last Modified By:    jvs


from dataclasses import dataclass
from typing import Sequence

from prometheus_client.metrics_core import GaugeMetricFamily, Metric # type: ignore

from ampel.core.AmpelDB import AmpelDB
from ampel.enum.DocumentCode import DocumentCode


@dataclass
class AmpelDBCollector:
    """
    Collect various metrics from the Ampel DB collections. These are quantities
    that are not implicitly collected by any worker process.
    """

    db: AmpelDB

    def collect(self) -> Sequence[Metric]:

        metrics = []
        try:
            metrics.append(
                GaugeMetricFamily(
                    "ampel_t2_docs_queued",
                    "Number of T2 docs awaiting processing",
                    value=self.db.get_collection("t2").count_documents(
                        {"code": DocumentCode.NEW}
                    ),
                )
            )
        except Exception:
            ...
        return metrics



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/aux/filter/AbsLogicOperatorFilter.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.01.2020
# Last Modified Date:  18.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Generic, Any
from collections.abc import Sequence
from ampel.types import T
from ampel.base.decorator import abstractmethod
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.FlatAnyOf import FlatAnyOf
from ampel.model.aux.FilterCriterion import FilterCriterion
from ampel.abstract.AbsApplicable import AbsApplicable


# mypy: disable-error-code = empty-body
class AbsLogicOperatorFilter(Generic[T], AbsApplicable, abstract=True):

	filters: FilterCriterion | FlatAnyOf[FilterCriterion] | AllOf[FilterCriterion]

	@staticmethod
	@abstractmethod
	def _apply_filter(args: Sequence[T], f: FilterCriterion) -> list[T]:
		...

	def apply(self, args: Sequence[T]) -> list[T]:

		if isinstance(self.filters, FilterCriterion):
			return self._apply_filter(args, self.filters)

		if isinstance(self.filters, AllOf):
			ret: Any = args
			for f in self.filters.all_of:
				ret = self._apply_filter(ret, f)
			return ret

		if isinstance(self.filters, FlatAnyOf):
			ret = []
			for f in self.filters.any_of:
				for el in self._apply_filter(args, f):
					if el not in ret:
						ret.append(el)
			return ret

		raise ValueError("Incorrect type provided for parameter 'filters'")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/T3BufferExporterUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.08.2022
# Last Modified Date:  17.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator
from ampel.types import UBson, T3Send
from ampel.cli.T3BufferExporterStager import TextExportOptions
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.struct.UnitResult import UnitResult
from ampel.view.SnapView import SnapView
from ampel.struct.T3Store import T3Store
from ampel.cli.export import txt_export, bin_export, get_fd


class T3BufferExporterUnit(AbsT3ReviewUnit[SnapView]):

	# None means stdout
	fd: None | str = None
	binary: bool = False
	verbose: bool = True
	id_mapper: None | AbsIdMapper = None
	txt_options: TextExportOptions = TextExportOptions()


	def process(self,
		gen: Generator[SnapView, T3Send, None],
		t3s: T3Store
	) -> UBson | UnitResult:

		fd, close_fd = get_fd(self.binary, self.fd, True)

		if self.binary:
			bin_export(
				fd, # type: ignore[arg-type]
				self.serialize_views(gen), self.id_mapper, close_fd,
				logger = self.logger if self.verbose else None
			)

		else:
			txt_export(
				fd, # type: ignore[arg-type]
				self.serialize_views(gen), self.id_mapper,
				**self.txt_options.dict(), close_fd = close_fd,
				logger = self.logger if self.verbose else None
			)
		
		return None


	def serialize_views(self,
		gen: Generator[SnapView, T3Send, None],
	) -> Generator[AmpelBuffer, None, None]:
		for sv in gen:
			if isinstance(sv, SnapView):
				yield sv.serialize() # type: ignore
			else:
				yield sv # type: ignore



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/ExpressionParser.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  13.08.2022
# Last Modified By:    jvs

import ast
from pathlib import Path
from typing import Any

class ExpressionParser(ast.NodeVisitor):

    @classmethod
    def evaluate(cls, expression: str, context: dict) -> str:
        """Evaluate Argo-style template expressions"""
        tree = ast.parse(expression, mode="eval")
        self = cls(context)
        self.visit(tree)
        if isinstance(self._value, str):
            return self._value
        elif isinstance(self._value, Path):
            return self._value.read_text()
        raise ValueError(f"unable to resolve {expression}")

    def __init__(self, context: dict):
        self._context = context
        self._value: Any = None
    
    def visit_Name(self, node: ast.Name) -> None:
        self.generic_visit(node)
        self._value = self._context.get(node.id)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        self.generic_visit(node)
        if isinstance(self._value, dict):
            self._value = self._value.get(node.attr)
    
    def visit_Subscript(self, node: ast.Subscript) -> Any:
        self.generic_visit(node)
        if isinstance(self._value, dict):
            self._value = self._value.get(node.value)
        elif isinstance(self._value, list) and isinstance(node, ast.Num):
            self._value = self._value[node.value]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/ThreadedViewGenerator.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                20.04.2021
# Last Modified Date:  26.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from multiprocessing import JoinableQueue
from typing import Generator
from ampel.abstract.AbsT3ReviewUnit import T, T3Send
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.t3.stage.BaseViewGenerator import BaseViewGenerator


class ThreadedViewGenerator(BaseViewGenerator[T]):
	"""
	Does not craft views but loads them from internal JoinableQueue and yields them
	"""

	def __init__(self, unit_name: str, queue: "JoinableQueue[T]", stock_updr: MongoStockUpdater) -> None:
		super().__init__(unit_name = unit_name, stock_updr = stock_updr)
		self.queue: "JoinableQueue[T]" = queue
		# ensure this generator's queue is consumed at most once 
		self._it = iter(queue.get, None)


	def __iter__(self) -> Generator[T, T3Send, None]:
		for view in self._it:
			self.stocks.append(view.id)
			try:
				yield view
			finally:
				# mark item as done even if the consumer raises an exception
				self.queue.task_done()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/SimpleViewGenerator.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                20.04.2021
# Last Modified Date:  09.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Iterable
from collections.abc import Generator
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.config.AmpelConfig import AmpelConfig
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.t3.stage.BaseViewGenerator import BaseViewGenerator, T, T3Send


class SimpleViewGenerator(BaseViewGenerator[T]):

	def __init__(self,
		unit: AbsT3ReviewUnit,
		buffers: Iterable[AmpelBuffer],
		stock_updr: MongoStockUpdater,
		config: AmpelConfig,
	) -> None:

		super().__init__(unit_name = unit.__class__.__name__, stock_updr = stock_updr)
		self.buffers = buffers
		self.View = unit._View
		# ensure this generator's queue is consumed at most once 
		self._it = iter(self.buffers)
		self._config = config

	def __iter__(self) -> Generator[T, T3Send, None]:
		View = self.View
		l = self.stocks
		for ab in self._it:
			l.append(ab['id'])
			yield View.of(ab, self._config, freeze=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/T3ThreadedStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.04.2021
# Last Modified Date:  14.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from itertools import islice
from multiprocessing import JoinableQueue
from multiprocessing.pool import ThreadPool, AsyncResult
from collections.abc import Generator, Iterable

from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.view.SnapView import SnapView
from ampel.struct.T3Store import T3Store
from ampel.content.T3Document import T3Document
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.t3.stage.T3BaseStager import T3BaseStager
from ampel.t3.stage.ThreadedViewGenerator import ThreadedViewGenerator


class T3ThreadedStager(T3BaseStager, abstract=True):
	"""
	Consumes chunk by chunk the input generator of ampel buffer,
	generates views and updates the view generator of t3 units
	via a JoinableQueue.
	chunk_size determines the max number of views held
	into memory at a given time.

                      chunk size
	                  <------->    chunk #2
	Ampel buffers:    a b c d e |
	Views unit 1:     A B C D E |
	Views unit 2:     A B C D E |
	Ampel buffers:              | f g h i j
	Views unit 1:               | F G H I J
	Views unit 2:               | F G H I J
	                 ---------------------->
	                         time
	"""


	def proceed_threaded(self,
		t3_units: list[AbsT3ReviewUnit],
		buf_gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> Generator[T3Document, None, None]:
		"""
		Execute the method 'process' of t3 units with views crafted using the provided buffer generator and t3 store,
		handle potential results of t3 unit.
		Note: code is not optimized for compactness but for execution speed
		"""

		ts = time()

		try:

			# Create and start T3 units "process(...)" threads (generator will block)
			with ThreadPool(processes=len(t3_units)) as pool:

				queues, generators, async_results = self.create_threaded_generators(pool, t3_units)

				# Optimize by potentially grouping units associated with the same view type
				qdict: dict[type, list[JoinableQueue]] = {}
				for unit in t3_units:
					if unit.__class__._View not in qdict:
						qdict[unit.__class__._View] = []
					qdict[unit.__class__._View].append(queues[unit])

				qv = queues.values()

				try:
				
					while (buffers := list(islice(buf_gen, self.chunk_size)) if self.chunk_size else buf_gen):

						self.put_views(buffers, qdict)

						# Join view queues before possibly processing next chunk
						# to ensure t3 units process views at a similar pace
						for q in qv:
							q.join()

					# Send sentinel to threaded view generators
					for q in qv:
						q.put(None) # type: ignore[arg-type]

					# Collect potential unit results
					for async_res, generator, t3_unit in zip(async_results, generators, t3_units):

						# potential T3Document to be included in the T3Document
						if (t3_unit_result := async_res.get()):
							if (x := self.handle_t3_result(t3_unit, t3_unit_result, t3s, generator.stocks, ts)):
								yield x

				except RuntimeError as e:
					if "StopIteration" in str(e):
						return None
					raise e

			self.flush(t3_units)

		except Exception as e:
			self.flush(t3_units)
			self.event_hdlr.handle_error(e, self.logger)


	def create_threaded_generators(self,
		pool: ThreadPool,
		t3_units: list[AbsT3ReviewUnit],
		t3s: None | T3Store = None
	) -> tuple[
		dict[AbsT3ReviewUnit, "JoinableQueue[SnapView]"],
		list[ThreadedViewGenerator],
		list[AsyncResult]
	]:
		"""
		Create and start T3 units "process(...)" threads (generator will block)
		"""

		queues: dict[AbsT3ReviewUnit, JoinableQueue[SnapView]] = {}
		generators: list[ThreadedViewGenerator] = []
		async_results: list[AsyncResult] = []

		for t3_unit in t3_units:
			queues[t3_unit] = JoinableQueue()
			generators.append(
				ThreadedViewGenerator(
					t3_unit.__class__.__name__,
					queues[t3_unit],
					self.stock_updr
				)
			)
			async_results.append(
				pool.apply_async(
					t3_unit.process,
					args=(generators[-1], t3s)
				)
			)

		return queues, generators, async_results


	def put_views(self, buffers: Iterable[AmpelBuffer], qdict: dict[type[SnapView], list[JoinableQueue]]) -> None:
		"""
		Note: code in here is not optimized for compactness but for execution speed
		"""

		conf = self.context.config

		# Simple case: all t3 units are associated with the same type of view
		if len(qdict) == 1:

			View = next(iter(qdict.keys()))
			qs = next(iter(qdict.values()))
			vo = View.of

			# In paranoia mode, we create a new view from the same buffer for each t3 unit
			if self.paranoia_level:
				for ab in buffers:
					for q in qs:
						q.put(vo(ab, conf))
			else:
				for ab in buffers:
					v = vo(ab, conf)
					for q in qs:
						q.put(v)

		# t3 units are associated with different type of view
		else:

			# Paranoia or say two units == two view types (non-optimizable)
			itms = qdict.items()
			if self.paranoia_level or all(len(x) == 1 for x in qdict.values()):
				for ab in buffers:
					for View, qs in itms:
						vo = View.of
						for q in qs:
							q.put(vo(ab, conf))

			# Optimize by potentially grouping units associated with the same view type
			else:
				for ab in buffers:
					for View, qs in itms:
						view = View.of(ab, conf)
						for q in qs:
							q.put(view)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/T3AggregatingStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                08.12.2021
# Last Modified Date:  16.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from ampel.base.AmpelBaseModel import AmpelBaseModel
from typing import Any
from collections.abc import Generator, Sequence

from ampel.types import UBson, OneOrMany
from ampel.t3.T3DocBuilder import T3DocBuilder
from ampel.struct.T3Store import T3Store
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.content.MetaRecord import MetaRecord
from ampel.util.mappings import get_by_json_path


class TargetModel(AmpelBaseModel):
	unit: None | str
	config: None | int | str
	code: None | int
	field: OneOrMany[str]


class T3AggregatingStager(AbsT3Stager, T3DocBuilder):
	"""
	Example:

	unit: T3AggregatingStager
	config:
	  t2:
	  - unit: T2NedTap
	    field: "data[0].*"
	  - unit: T2NedSNCosmo
	  field:
	  - "data[0].fit_results"
	  - "data[0].covariance"

	will create a new t3 doc in the DB, whose body will contain the aggregated results.
	The doc could look like below (note that ampel ids were stringified to comply with BSON requirements):

	{
	  ...
	  'body': {
	    "33876" : {
	       "prefname" : "WISEA J235434.02+154441.8",
	       "pretype" : "G",
	       "ra" : 358.6417460104,
	       "dec" : 15.7449859573,
	       ...
	       "fit_results" : {
	         "z" : 0.07458627,
	         "t0" : 2459226.42052562,
	         "x0" : 0.000359877303537212,
	         ...
	       },
	       "covariance": [
	         [0.525561894902874, 5.08414469979e-06, 0.103650196867142, -0.0141356930137254],
	         ...
	       ],
	       ...
	   },
	   "496964" : {
	     "prefname" : "WISEA J114639.05+421201.3",
	     "pretype" : "G",
	     "ra" : 176.6627824999,
	     "dec" : 42.2004549186,
	     ...
	     "fit_results" : {
	       "z" : 0.0512447,
	       "t0" : 2459227.16179618,
	       "x0" : 0.000925982418211906,
	       ...
	     },
	     "covariance" : [
	       [0.0427179910190306, 7.18629916088547e-07, 0.000100038844775205, -0.000266005925059596],
	       ...
	     ],
	     ...
	  }
	...
	"""

	# Override
	save_stock_ids: bool = True

	#: Only applies to doc output
	split_tiers: bool = False

	t0: None | OneOrMany[TargetModel]
	t1: None | OneOrMany[TargetModel]
	t2: None | OneOrMany[TargetModel]


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> None | Generator[T3Document, None, None]:

		t0 = [self.t0] if isinstance(self.t0, TargetModel) else self.t0
		t1 = [self.t1] if isinstance(self.t1, TargetModel) else self.t1
		t2 = [self.t2] if isinstance(self.t2, TargetModel) else self.t2

		if t0:
			t0d: dict[str, Any] = {}

		if t1:
			t1d: dict[str, Any] = {}

		if t2:
			t2d: dict[str, Any] = {}

		for ab in gen:

			sid = str(ab['id'])

			if t0:
				raise NotImplementedError("Please implement")

			if t1:
				raise NotImplementedError("Please implement")

			if t2:
				for model in t2:
					for t2doc in ab.get('t2') or []:
						if t2doc['unit'] != model.unit:
							continue
						if model.code is not None and t2doc['code'] != model.code:
							continue
						if body := self.get_t2_payload(t2doc['body'], t2doc['meta'], model.code):
							if model.field:
								for f in [model.field] if isinstance(model.field, str) else model.field:
									if (ret := get_by_json_path(body, f)):
										if sid not in t2d:
											t2d[sid] = {}
										if ret[0] == "*":
											t2d[sid].update(ret[1])
										else:
											t2d[sid][ret[0]] = ret[1]
							else:
								t2d[sid].update(ret)

		if t0 and not t1 and not t2:
			return self._craft(t0d, 't0', t3s)

		if t1 and not t0 and not t2:
			return self._craft(t1d, 't1', t3s)

		if t2 and not t1 and not t0:
			return self._craft(t2d, 't2', t3s)

		out: dict[str, Any] = {}

		if t0:
			self._upd(out, t0d, 't0')
		if t1:
			self._upd(out, t1d, 't1')
		if t2:
			self._upd(out, t2d, 't2')

		return self._craft(out, '', t3s)


	def _craft(self, d: dict[str, Any], s: str, t3s: T3Store) -> Generator[T3Document, None, None]:
		yield self.craft_t3_doc(
			self, # type: ignore
			{k: {s: v} for k, v in d.items()} if self.split_tiers else d,
			t3s,
			time(),
			[int(el) for el in d.keys()]
		)


	def _upd(self, out: dict[str, Any], d: dict[str, Any], s: str) -> None:

		if self.split_tiers:
			for k, v in d.items():
				out[k] = {s: v}
		else:
			for k, v in d.items():
				if k in out:
					out[k].update(v)
				else:
					out[k] = v


	def get_t2_payload(self,
		body: None | Sequence[UBson],
		meta: Sequence[MetaRecord],
		code: None | int = None
	) -> None | dict[str, Any]:
		"""
		:returns: the content of the last array element of body associated with a meta code >= 0 or equals code arg.
		"""
		if not body:
			return None

		idx = len(
			[
				el for el in meta
				if el['tier'] == 2 and
				(el['code'] >= 0 if code is None else el['code'] == code)
			]
		) - 1

		if idx == -1:
			return None

		# A manual/admin $unset: {body: 1} was used to delete bad data
		if idx > len(body) - 1:
			idx = len(body) - 1

		if idx >= 0 and isinstance(body[idx], dict):
			return body[idx] # type: ignore[return-value] # remove when mypy gets smarter

		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/EventHandler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.09.2018
# Last Modified Date:  19.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from bson import ObjectId
from typing import Any, Literal, TYPE_CHECKING
from ampel.enum.EventCode import EventCode
from ampel.struct.Resource import Resource
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.AmpelLoggingError import AmpelLoggingError
from ampel.log.utils import report_exception

if TYPE_CHECKING:
	from ampel.core.AmpelDB import AmpelDB


# TODO: (much later) remove explicit dependency on pymongo
class EventHandler:
	"""
	Handles the creation and publication of event documents into the event database
	"""

	def __init__(self,
		process_name: str,
		ampel_db: 'AmpelDB',
		col_name: str = "event",
		raise_exc = False,
		dry_run: bool = False,
		job_sig: None | int = None,
		resources: None | dict[str, Resource] = None,
		extra: None | dict[str, Any] = None
	):
		"""
		:param col_name: name of db collection to use (default 'events').
		"""
		self.process_name = process_name
		self.db = ampel_db
		self.raise_exc = raise_exc
		self.dry_run = dry_run
		self.job_sig = job_sig
		self.dry_run = dry_run
		self.run_id: None | int = None
		self.code: None | EventCode = None
		self.col = ampel_db.get_collection(col_name)
		self.extra: dict[str, Any] = extra or {}
		self.resources: dict[str, Resource] = resources or {}


	def register(self,
		run_id: None | int = None,
		tier: None | Literal[-1, 0, 1, 2, 3] = None,
		task_nbr: None | int = None,
		code: EventCode = EventCode.RUNNING
	) -> None:

		doc = {'process': self.process_name} | self.extra
		doc['code'] = code

		if tier:
			doc['tier'] = tier

		if run_id is not None:
			self.set_run_id(run_id)
			doc['run'] = run_id

		if self.job_sig:
			doc['jobid'] = self.job_sig

		self.dkeys = doc.keys()
		self.extra = {}

		if self.dry_run:
			self.ins_id = ObjectId()
		else:
			self.ins_id = self.col.insert_one(doc).inserted_id # type: ignore[arg-type]


	def set_run_id(self, val: int) -> None:
		if self.run_id is not None:
			raise ValueError("run id already set")
		self.run_id = val


	def get_run_id(self) -> int:
		if self.run_id is None:
			raise ValueError("run id not set")
		return self.run_id


	def add_extra(self, overwrite: bool = False, logger: None | AmpelLogger = None, **extra) -> None:
		for k, v in extra.items():
			if (k in self.extra or k in self.dkeys) and not overwrite:
				if logger:
					logger.error(f"Cannot overwrite already existing event value for key {k}")
				continue
			self.extra[k] = v


	def add_resource(self, resource: Resource, overwrite: bool = False) -> None:
		""" save references to resources generated dynamically by t3 units """
		if self.resources and not overwrite and resource.name in self.resources:
			raise ValueError(
				f"Resource name '{resource.name}' already defined"
				f"(use overwrite=True to ignore)"
			)
		self.resources[resource.name] = resource


	def set_tier(self, val: Literal[0, 1, 2, 3]) -> None:
		self.extra['tier'] = val


	def set_code(self, val: EventCode):
		if self.code == EventCode.EXCEPTION and val != EventCode.EXCEPTION:
			raise ValueError("Cannot override EventCode.EXCEPTION")
		self.code = val


	def handle_error(self, e: Exception, logger: AmpelLogger) -> None:

		self.code = EventCode.EXCEPTION

		if self.raise_exc:
			raise e

		# Try to insert doc into trouble collection (raises no exception)
		# Exception will be logged out to console
		report_exception(
			self.db, logger, exc=e,
			info={'process': self.process_name}
		)


	def update(self, _save_duration: bool = True) -> None:
		""" :raises: AmpelLoggingError """

		upd: dict[str, Any] = self.extra.copy()
		if self.code is not None:
			upd['code'] = self.code

		if _save_duration:
			upd['duration'] = round(time() - self.ins_id.generation_time.timestamp(), 3)

		if not upd or self.dry_run:
			return

		res = self.col.update_one( # type: ignore[union-attr]
			{'_id': self.ins_id},
			{'$set': upd}
		)

		if res.modified_count == 0 and res.upserted_id is None:
			raise AmpelLoggingError(
				"Events collection update failed (%s)" % {
					'mongoUpdateResult': res.raw_result,
					'process': self.process_name
				}
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/db/FieldModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                19.10.2019
# Last Modified Date:  13.04.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelBaseModel import AmpelBaseModel

class FieldModel(AmpelBaseModel):

	field: str
	direction: int = 1

	def dict(self, **kwargs):
		return (self.field, self.direction)

	def get_id(self) -> str:
		"""
		Returns an indexId similar to what pymongo index_information outputs.
		Ex: [('tranId', 1), ('channel', 1)] -> tranId_1_channel_1
		"""
		# Shortcut
		return f"{self.field}_{self.direction}"



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# Modifications for Ampel:
# - removed Python <= 3.8 compat + dependencies on talisker itself
# - removed locks (only ever called from one process)
# - added implicit per-worker labels

# -*- coding: utf-8 -*-

import glob
import os
import tempfile
from typing import Collection

from prometheus_client import (
    Metric,
    CollectorRegistry,
    core,
    generate_latest,
    mmap_dict,
    multiprocess,
)

histogram_archive = "histogram_archive.db"
counter_archive = "counter_archive.db"


def collect_metrics():

    if "PROMETHEUS_MULTIPROC_DIR" in os.environ:
        registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)
    else:
        registry = core.REGISTRY
    return generate_latest(registry)


def prometheus_setup_worker(labels: None | dict[str,str] = None) -> None:
    """
    Monkey-patch mmap_key and ValueClass to add implicit labels. This must be
    done before any metrics are instantiated.
    """
    if labels is not None:
        from prometheus_client import values

        def mmap_key(metric_name: str, name: str, labelnames: list[str], labelvalues: list[str], help_text: str) -> str:
            return mmap_dict.mmap_key(
                metric_name,
                name,
                list(labels.keys()) + list(labelnames) if labels else labelnames,
                list(labels.values()) + list(labelvalues) if labels else labelvalues,
                help_text,
            )

        values.mmap_key = mmap_key
        # synthesize a new ValueClass (captures mmap_key)
        values.ValueClass = values.get_value_class()


def prometheus_cleanup_worker(pid: int) -> None:
    """
    Aggregate dead worker's metrics into a single archive file, preventing
    collection time from growing without bound as pointed out in
    - https://github.com/prometheus/client_python/issues/568
    - https://github.com/prometheus/client_python/issues/443
    - https://github.com/prometheus/client_python/pull/430
    - https://github.com/prometheus/client_python/pull/441
    """

    multiprocess.mark_process_dead(pid)  # this takes care of gauges
    prom_dir = os.environ["PROMETHEUS_MULTIPROC_DIR"]
    # also remove max gauges
    for f in glob.glob(os.path.join(prom_dir, f"gauge_max_{pid}.db")):
        os.remove(f)

    # check at least one worker file exists
    if not (
        paths := [
            worker_file
            for kind in ("histogram", "counter")
            if os.path.exists(worker_file := os.path.join(prom_dir, f"{kind}_{pid}.db"))
        ]
    ):
        return

    histogram_path = os.path.join(prom_dir, histogram_archive)
    counter_path = os.path.join(prom_dir, counter_archive)
    archive_paths = [p for p in [histogram_path, counter_path] if os.path.exists(p)]

    collect_paths = paths + archive_paths
    collector = multiprocess.MultiProcessCollector(None)

    metrics: Collection[Metric] = collector.merge(collect_paths, accumulate=False)

    tmp_histogram = tempfile.NamedTemporaryFile(delete=False)
    tmp_counter = tempfile.NamedTemporaryFile(delete=False)
    write_metrics(metrics, tmp_histogram.name, tmp_counter.name)

    # no lock here, since this is only ever called from the asyncio event loop
    # of a single process
    os.rename(tmp_histogram.name, histogram_path)
    os.rename(tmp_counter.name, counter_path)

    for path in paths:
        os.unlink(path)


def write_metrics(metrics: Collection[Metric], histogram_file: str, counter_file: str) -> None:

    histograms = mmap_dict.MmapedDict(histogram_file)
    counters = mmap_dict.MmapedDict(counter_file)

    try:
        for metric in metrics:
            if metric.type == "histogram":
                sink = histograms
            elif metric.type == "counter":
                sink = counters
            else:
                continue

            for sample in metric.samples:
                # prometheus_client 0.4+ adds extra fields
                name, labels, value = sample[:3]
                key = mmap_dict.mmap_key(
                    metric.name, name, list(labels.keys()), list(labels.values()), metric.documentation,
                )
                sink.write_value(key, value)
    finally:
        histograms.close()
        counters.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/complement/T3LogsAppender.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.03.2021
# Last Modified Date:  28.02.2022
# Last Modified By:    Marcus Fenner <mf@physik.hu-berlin.de>

from datetime import datetime
from bson.objectid import ObjectId
from typing import Any
from collections.abc import Iterable
from ampel.struct.T3Store import T3Store
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.mongo.query.var.LogsLoader import LogsLoader
from ampel.log.utils import safe_query_dict
from ampel.abstract.AbsBufferComplement import AbsBufferComplement


class T3LogsAppender(AbsBufferComplement):

	use_last_run: bool = True
	logs_loader_conf: dict[str, Any] = {}

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.log_loader = LogsLoader(**self.logs_loader_conf, read_only=True)
		self.col = self.context.db.get_collection('log')


	def complement(self, it: Iterable[AmpelBuffer], t3s: T3Store) -> None:

		query: dict[str, Any] = {'s': {'$in': [el['id'] for el in it]}}

		if t3s.session and self.use_last_run and t3s.session.get('last_run'):
			query['_id'] = {
				'$gte': ObjectId.from_datetime(
					datetime.utcfromtimestamp(
						t3s.session['last_run']
					)
				)
			}

		logs = list(
			self.log_loader.fetch_logs(self.col, query)
		)

		self.logger.debug(
			f"Log query returned {len(logs)} result(s)",
			extra=safe_query_dict(query)
		)

		if not logs:
			return

		for ab in it:
			if 'logs' not in ab or ab['logs'] is None:
				ab['logs'] = []
			for l in logs:
				if ab['id'] == l['s']:
					ab['logs'].append(l) # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/T3Processor.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.02.2018
# Last Modified Date:  19.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from importlib import import_module
from typing import Any, Annotated

from ampel.types import OneOrMany
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.abstract.AbsT3ControlUnit import AbsT3ControlUnit
from ampel.abstract.AbsProcessorTemplate import AbsProcessorTemplate
from ampel.struct.T3Store import T3Store
from ampel.view.T3DocView import T3DocView
from ampel.model.UnitModel import UnitModel
from ampel.model.t3.T3IncludeDirective import T3IncludeDirective
from ampel.core.EventHandler import EventHandler
from ampel.log import AmpelLogger, LogFlag, SHOUT


class T3Processor(AbsEventUnit):

	template: None | str = None
	include: None | T3IncludeDirective
	execute: Annotated[OneOrMany[UnitModel], AbsT3ControlUnit]


	def __init__(self, **kwargs) -> None:

		if 'template' in kwargs:
			tpl_name = kwargs.pop("template")
			ctx = kwargs.pop("context")
			if ctx is None:
				raise ValueError("Context required")

			if tpl_name not in ctx.config._config.get('template', []):
				raise ValueError(f"Unknown process template: {tpl_name}")

			fqn = ctx.config._config['template'][tpl_name]
			class_name = fqn.split(".")[-1]
			Tpl = getattr(import_module(fqn), class_name)
			if not issubclass(Tpl, AbsProcessorTemplate):
				raise ValueError(f"Unexpected template type: {Tpl}")

			tpl = Tpl(
				**{
					k: v for k, v in kwargs.items()
					if k not in AbsEventUnit._annots
				}
			)
			kwargs.update(
				tpl.get_model(ctx.config._config, kwargs).dict()['config']
			)
			kwargs['context'] = ctx

		super().__init__(**kwargs)


	def proceed(self, event_hdlr: EventHandler) -> None:

		event_hdlr.set_tier(3)
		logger = AmpelLogger.from_profile(
			self.context, self.log_profile, event_hdlr.get_run_id(),
			base_flag = LogFlag.T3 | LogFlag.CORE | self.base_log_flag,
			force_refresh = True
		)

		try:

			# Feedback
			logger.log(SHOUT, f'Running {self.process_name}')

			loader = self.context.loader
			t3s = T3Store()

			if self.include:

				if self.include.docs:
					pass # later

				if (x := self.include.session):
					sdict: dict[str, Any] = {}
					for model in [x] if isinstance(x, UnitModel) else x:
						rd = loader.new_context_unit(
							model = model,
							context = self.context,
							sub_type = AbsT3Supplier,
							event_hdlr = event_hdlr,
							logger = logger
						).supply(t3s)
						if rd:
							sdict |= rd
					if sdict:
						t3s.add_session_info(sdict)

			units = self.context.config._config['unit']
			for i, um in enumerate([self.execute] if isinstance(self.execute, UnitModel) else self.execute):

				if um.unit not in units:
					raise ValueError(f"Unknown unit: {um.unit}")

				if "AbsT3ControlUnit" not in units[um.unit]['base']:
					raise ValueError("T3Processor executes only AbsT3ControlUnit units")

				t3_unit = loader.new_context_unit(
					model = um,
					context = self.context,
					sub_type = AbsT3ControlUnit,
					logger = logger,
					event_hdlr = event_hdlr,
					channel = self.channel
				)

				logger.info(
					f"Processing run block {i}",
					extra={'unit': t3_unit.__class__.__name__}
				)

				# Potential T3Document to be included in the t3 collection
				if (ret := t3_unit.process(t3s)):
					for t3d in ret:
						t3s.add_view(T3DocView.of(t3d, self.context.config))
						if 'meta' not in t3d:
							raise ValueError("Invalid T3Document")
						t3d['meta']['traceid'] = {'t3processor': self._trace_id}
						if event_hdlr.job_sig:
							t3d['meta']['jobid'] = event_hdlr.job_sig
						self.context.db.get_collection('t3').insert_one(t3d) # type: ignore[arg-type]

			if t3s.resources:
				for v in t3s.resources.values():
					event_hdlr.add_resource(v)

		except Exception as e:
			event_hdlr.handle_error(e, logger)

		# Feedback
		logger.log(SHOUT, f'Done running {self.process_name}')
		logger.flush()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/unit/T3PrintStore.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.09.2022
# Last Modified Date:  26.09.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.struct.T3Store import T3Store
from ampel.abstract.AbsT3PlainUnit import AbsT3PlainUnit


class T3PrintStore(AbsT3PlainUnit):

	# Otherwise logger will be used
	do_print: bool = True

	def process(self, t3s: None | T3Store = None):

		p: Any = print if self.do_print else self.logger.info
		p(f"Running {self.__class__.__name__}")

		if not t3s:
			p("No T3 store available")
			return

		p(f"Units: {t3s.units}")

		if t3s.views:
			# To be implemented
			pass

		if t3s.session:
			# To be implemented
			pass

		if t3s.extra:
			p(f"Extra: {t3s.extra}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/SimpleT2BasedSupplier.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.07.2021
# Last Modified Date:  15.07.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from collections.abc import Generator
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.log.utils import safe_query_dict
from ampel.struct.T3Store import T3Store


class SimpleT2BasedSupplier(AbsT3Supplier):

	query: dict[str, Any]

	#: minimum # of t2 docs per stock (useful in combination with $or queries)
	min_docs: None | int

	def supply(self, t3s: T3Store) -> Generator[AmpelBuffer, None, None]:

		if self.logger.verbose > 1: # log query parameters
			self.logger.debug(
				None, extra={
					'col': 2,
					'query': safe_query_dict(self.query, dict_key=None)
				}
			)

		# Retrieve pymongo cursor
		col = self.context.db.get_collection("t2")

		d: dict[int, AmpelBuffer] = {}
		for el in col.find(self.query):
			if el['stock'] in d:
				d[el['stock']]['t2'].append(el) # type: ignore
			else:
				d[el['stock']] = AmpelBuffer(
					id = el['stock'],
					t2 = [el]
				)

		if (md := self.min_docs):
			for ab in d.values():
				if ab['t2'] is None or len(ab['t2']) < md:
					continue
				yield ab
		else:
			for ab in d.values():
				yield ab



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/MongoT2Ingester.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.12.2017
# Last Modified Date:  08.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from pymongo import UpdateOne
from typing import Any
from ampel.enum.DocumentCode import DocumentCode
from ampel.content.T2Document import T2Document
from ampel.mongo.utils import maybe_use_each
from ampel.abstract.AbsDocIngester import AbsDocIngester


class MongoT2Ingester(AbsDocIngester[T2Document]):

	def ingest(self, doc: T2Document) -> None:

		# Note: mongodb $setOnInsert does not retain key order
		set_on_insert: dict[str, Any] = {'code': DocumentCode.NEW.value}
		add_to_set: dict[str, Any] = {'channel': maybe_use_each(doc['channel'])}

		match = {
			'stock': doc['stock'],
			'unit': doc['unit'],
			'config': doc['config'],
			'link': doc['link']
		}

		if 'origin' in doc:
			match['origin'] = doc['origin']

		if 'col' in doc:
			set_on_insert['col'] = doc['col']

		if 'tag' in doc:
			add_to_set['tag'] = maybe_use_each(doc['tag'])

		# Append update operation to bulk list
		self.updates_buffer.add_t2_update(
			UpdateOne(
				match,
				{
					'$setOnInsert': set_on_insert,
					'$addToSet': add_to_set,
					'$push': {'meta': maybe_use_each(doc['meta'])} # meta must be set by compiler
				},
				upsert=True
			)
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/ChannelConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  02.01.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.util.hash import hash_payload
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class ChannelConfigCollector(AbsDictConfigCollector):

	def add(self,
		chan_dict: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:

		if 'channel' not in chan_dict:
			self.missing_key(
				what='channel', key='channel',
				dist_name=dist_name, register_file=register_file
			)
			return

		try:

			chan_name = chan_dict['channel']

			if self.verbose:
				self.logger.log(VERBOSE, f'Adding channel: {chan_name}')

			# Check duplicated channel names between distribs
			if self.check_duplicates(chan_name, dist_name, version, register_file):
				return

			if not ('NO_HASH' in chan_dict.get('policy', []) or isinstance(chan_name, int)):
				chan_dict['hash'] = hash_payload(chan_name)
				for k, v in self.items():
					if chan_dict['hash'] == v.get('hash'):
						raise ValueError(
							f'Channel name hash collision detected. '
							f'Channel name 1: {k}. Hash value: {v.get("hash")}'
							f'Channel name 2: {chan_dict["channel"]}. Hash value: {chan_dict.get("hash")}'
						)

			self.__setitem__(chan_name, chan_dict)

		except Exception as e:
			self.error(
				f'Error occured while loading channel config {self.distrib_hint(dist_name, register_file)}. '
				f'Offending value: {chan_dict}',
				exc_info=e
			)



import copy, pytest
from ampel.config.builder.ConfigValidator import ConfigValidator
from ampel.template.PeriodicSummaryT3 import PeriodicSummaryT3


@pytest.mark.parametrize("loader_directives", [None, [{"col": "stock"}], ["TRANSIENT"]])
def test_validate(core_config, loader_directives, ampel_logger):
    """
    PeriodicSummaryT3 emits a valid process.
    """
    config = copy.deepcopy(core_config)
    config["process"]["t3"]["foo"] = (
        PeriodicSummaryT3(
            **{
                "name": "foo",
                "schedule": "every().day.at('15:00')",
                "channel": "FOO",
                "load": loader_directives,
                "run": {"unit": "DemoReviewT3Unit"},
            }
        )
        .get_process(config, ampel_logger) | {"version": 0}
    )
    # from ampel.core.EventHandler import EventHandler
    # eh = EventHandler("foo", ampel_db=None, tier=3, dry_run=True)
    assert ConfigValidator(config).validate() == config

def test_single_element_run_sequence(core_config, ampel_logger):
    config = copy.deepcopy(core_config)
    config["process"]["t3"]["foo"] = (
        PeriodicSummaryT3(
            **{
                "name": "foo",
                "schedule": "every().day.at('15:00')",
                "channel": {"any_of": ["HU_GP_10", "HU_GP_59"]},
                "load": ["TRANSIENT", "DATAPOINT", "T2RECORD"],
                "run": [{"unit": "DemoReviewT3Unit", "config": {}}],
            }
        )
        .get_process(config, ampel_logger) | {"version": 0}
    )
    assert ConfigValidator(config).validate() == config
    assert config["process"]["t3"]["foo"]["channel"] is None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/AbsCoreCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.03.2021
# Last Modified Date:  11.01.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json
import re, os
from typing import Any, TypeVar, Literal
from collections.abc import Sequence, Iterator
from ampel.config.AmpelConfig import AmpelConfig
from ampel.abstract.AbsCLIOperation import AbsCLIOperation
from ampel.core.AmpelContext import AmpelContext
from ampel.core.UnitLoader import UnitLoader
from ampel.log.AmpelLogger import AmpelLogger
from ampel.util.mappings import set_by_path
from ampel.util.freeze import recursive_freeze
from ampel.util.pretty import out_stack
from ampel.cli.utils import get_vault, get_db, _maybe_int
from ampel.cli.config import get_user_data_config_path

custom_conf_patter = re.compile(r"^--[\w-]*(?:\.[\w-]+)*.*$")

T = TypeVar("T", bound = AmpelContext)

class AbsCoreCommand(AbsCLIOperation, abstract=True):

	def __init__(self):
		self.parsers = {}
		try:
			import sys, IPython
			sys.breakpointhook = IPython.embed
		except Exception:
			pass


	def load_config(self,
		config_path: None | str,
		unknown_args: Sequence[str],
		logger: None | AmpelLogger = None,
		freeze: bool = True,
		env_var_prefix: None | str = "AMPEL_CONFIG_",
	) -> AmpelConfig:

		if not config_path:
			std_conf = get_user_data_config_path()
			if os.path.exists(std_conf):
				ampel_conf = AmpelConfig.load(std_conf, freeze=False)
			else:
				with out_stack():
					raise ValueError("No default ampel config found -> argument -config required\n")
		else:
			ampel_conf = AmpelConfig.load(config_path, freeze=False)

		if logger is None:
			logger = AmpelLogger.get_logger()

		if env_var_prefix is not None:
			for var, value in os.environ.items():
				if var.startswith(env_var_prefix):
					k = var[len(env_var_prefix):]
					try:
						v = json.loads(value)
					except json.JSONDecodeError:
						v = value
					logger.info(f"Setting config parameter '{k}' from environment")
					set_by_path(ampel_conf._config, k, v)

		for k, v in self.get_custom_args(unknown_args):

			if ampel_conf.get(".".join(k.split(".")[:-1])) is None:
				with out_stack():
					raise ValueError(f"Unknown config parameter '{k}'\n")

			logger.info(f"Setting config parameter '{k}' value to: {v}")
			set_by_path(ampel_conf._config, k, v)

		if freeze:
			ampel_conf._config = recursive_freeze(ampel_conf._config)

		return ampel_conf


	def get_custom_args(self, customizations: Sequence[str]) -> Iterator[tuple[str, Any]]:
		"""
		:raises: ValueError
		"""

		it = iter(customizations)
		for el in it:
			if custom_conf_patter.match(el):
				if "=" in el:
					s = el.split("=")
					k = s[0][2:]
					v = _maybe_int(s[1])
				else:
					k = el[2:]
					try:
						v = _maybe_int(next(it))
					except StopIteration:
						v = None
				yield k, v
			else:
				with out_stack():
					raise ValueError(f"Unknown argument: {el}\n")


	def get_context(self,
		args: dict[str, Any],
		unknown_args: Sequence[str],
		logger: None | AmpelLogger = None,
		freeze_config: bool = True,
		ContextClass: type[T] = AmpelContext, # type: ignore[assignment]
		require_existing_db: bool | str = True,
		one_db: bool | Literal['auto'] = False,
		**kwargs
	) -> T:
		"""
		:require_existing_db: str typed values specify required database prefix
		"""

		if logger is None:
			logger = AmpelLogger.get_logger()

		config = self.load_config(
			args['config'], unknown_args, logger, freeze = freeze_config
		)

		vault = get_vault(args)
		db = get_db(config, vault, require_existing_db, one_db)
		return ContextClass(
			config = config,
			db = db,
			loader = UnitLoader(config, db=db, vault=vault),
			**kwargs
		)


	def convert_logical_args(self, name: str, args: dict[str, Any]) -> None:

		for k in (f"with_{name}", f"with_{name}s_and", f"with_{name}s_or"):
			if args.get(k):
				if name not in args:
					args[name] = {}
				args[name]['with'] = args[k]
				break

		for k in (f"without_{name}", f"without_{name}s_and", f"without_{name}s_or"):
			if args.get(k):
				if name not in args:
					args[name] = {}
				args[name]['without'] = args[k]
				break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/demo/DemoResourceT3Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                19.12.2022
# Last Modified Date:  19.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from random import randint
from ampel.types import UBson
from ampel.struct.T3Store import T3Store
from ampel.struct.Resource import Resource
from ampel.struct.UnitResult import UnitResult
from ampel.abstract.AbsT3PlainUnit import AbsT3PlainUnit


class DemoResourceT3Unit(AbsT3PlainUnit):

	debug: bool = False

	def post_init(self) -> None:
		self.logger.info("post_init was called")

	def process(self, t3s: T3Store) -> UBson | UnitResult:
		self.logger.info("Running DemoResourceT3Unit")
		r = Resource(name='demoToken', value=randint(10000000000000, 90000000000000))
		t3s.add_resource(r)
		return UnitResult(body=r.dict()) if self.debug else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/abstract/AbsProcessTemplate.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.10.2019
# Last Modified Date:  27.10.2019
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.log.AmpelLogger import AmpelLogger
from ampel.base.decorator import abstractmethod
from ampel.config.builder.AbsConfigTemplate import AbsConfigTemplate


class AbsProcessTemplate(AbsConfigTemplate, abstract=True):

	@abstractmethod
	def get_process(self, config: dict[str, Any], logger: AmpelLogger) -> dict[str, Any]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/AmpelDB.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.06.2018
# Last Modified Date: 27.08.2022
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import secrets, collections.abc
from typing import Any, Literal
from functools import cached_property
from collections import defaultdict  # type: ignore[attr-defined]
from collections.abc import Sequence
from pymongo import MongoClient, ReadPreference
from pymongo.database import Database
from pymongo.collection import Collection
from pymongo.errors import ConfigurationError, DuplicateKeyError, OperationFailure, CollectionInvalid

from ampel.types import ChannelId
from ampel.mongo.utils import get_ids
from ampel.log.AmpelLogger import AmpelLogger
from ampel.config.AmpelConfig import AmpelConfig
from ampel.base.AmpelUnit import AmpelUnit
from ampel.secret.AmpelVault import AmpelVault
from ampel.mongo.view.AbsMongoView import AbsMongoView
from ampel.mongo.view.MongoOneView import MongoOneView
from ampel.mongo.view.MongoOrView import MongoOrView
from ampel.mongo.view.MongoAndView import MongoAndView
from ampel.mongo.model.AmpelColModel import AmpelColModel
from ampel.mongo.model.AmpelDBModel import AmpelDBModel
from ampel.mongo.model.IndexModel import IndexModel
from ampel.mongo.model.ShortIndexModel import ShortIndexModel
from ampel.mongo.model.MongoClientOptionsModel import MongoClientOptionsModel
from ampel.mongo.model.MongoClientRoleModel import MongoClientRoleModel
from ampel.util.collections import try_reduce

intcol = {'t0': 0, 't1': 1, 't2': 2, 't3': 3, 'stock': 4}


class UnknownDatabase(Exception):
	pass


class AmpelDB(AmpelUnit):
	"""
	Ampel stores information in a dedicated DB.
	This class allows to create or retrieve the underlying database collections.
	"""

	prefix: str = 'Ampel'
	databases: Sequence[AmpelDBModel]
	mongo_uri: str
	mongo_options: MongoClientOptionsModel = MongoClientOptionsModel()
	vault: None | AmpelVault
	require_exists: bool = False
	one_db: bool = False
	#: Route reads to secondaries of a replica set. This can increase
	#: performance at the expense of consistency
	read_from_secondary: bool = True


	@classmethod
	def new(cls,
		config: AmpelConfig,
		vault: None | AmpelVault = None,
		require_exists: bool | str = False,
		one_db: bool | Literal['auto'] = False
	) -> 'AmpelDB':
		"""
		:param require_exists:
		- False: AmpelDB instantiation will succeed even if underlying database(s) do not exist
		- True: AmpelDB instantiation will fail unless underlying databases exist
		- str: same as True but provided value overrides prefix value otherwise loaded from AmpelConfig
		:param one_db: whether to create or load a single database rather than having
		collections split in in three databases (data, var, ext).
		If 'auto' is set (for which require_exists must be True), then AmpelDB will try
		to load a databse in single mode and try again in multi-mode if it fails.
		Note that ampel 'jobs' usually operate with one_db=True.
		:raises: ValueError in case a required config entry is missing
		"""
		if one_db == 'auto':
			if require_exists is False:
				raise ValueError("Option 'one_db' cannot be set to 'auto' if require_exists is False")
			try:
				return cls._new(config, vault, require_exists, True)
			except UnknownDatabase:
				return cls._new(config, vault, require_exists, False)
		return cls._new(config, vault, require_exists, one_db)


	@classmethod
	def _new(cls,
		config: AmpelConfig,
		vault: None | AmpelVault = None,
		require_exists: bool | str = False,
		one_db: bool = False,
	) -> 'AmpelDB':
		""" :raises: ValueError in case a required config entry is missing """
		db_config = config.get('mongo', dict, raise_exc=True)
		if isinstance(require_exists, str):
			dict.__setitem__(db_config, 'prefix', require_exists)
			require_exists = True
		return cls(
			mongo_uri = config.get('resource.mongo', str, raise_exc=True),
			vault = vault,
			require_exists = require_exists,
			one_db = one_db,
			**db_config
		)


	def __init__(self, **kwargs) -> None:

		if 'ingest' in kwargs:
			kwargs.pop('ingest')

		super().__init__(**kwargs) # type: ignore[call-arg]

		self.col_config: dict[str, AmpelColModel] = {
			col.name: col
			for db_config in self.databases
			for col in db_config.collections
		}

		self.mongo_collections: dict[str, dict[str, Collection]] = {}
		self.mongo_clients: dict[str, MongoClient] = {} # map role with client

		if self.require_exists and not self._get_pymongo_db("data", role="w").list_collection_names():
			raise UnknownDatabase(f"Database(s) with prefix {self.prefix} do not exist")


	@cached_property
	def col_trace_ids(self) -> Collection:
		return self.get_collection('trace')
	
	@cached_property
	def col_conf_ids(self) -> Collection:
		return self.get_collection('conf')

	@cached_property
	def trace_ids(self) -> set[int]:
		return get_ids(self.col_trace_ids)
	
	@cached_property
	def conf_ids(self) -> set[int]:
		return get_ids(self.col_conf_ids)


	def enable_rejected_collections(self, channel_names: Sequence[ChannelId]) -> None:
		"""
		Makes rejected collections (DB: Ampel_rej, collection: <channel_name>)
		available through standard method call AmpelDB.get_collection(channel_name)
		:param channel_names: list of channel names
		"""
		db_config = AmpelDBModel(
			name = 'rej',
			collections = [
				AmpelColModel(
					name = chan_name if isinstance(chan_name, str) else str(chan_name)
				)
				for chan_name in channel_names
			],
			role = MongoClientRoleModel(r='logger', w='logger')
		)

		self.databases = list(self.databases) + [db_config]

		for col in db_config.collections:
			self.col_config[col.name] = col


	def get_collection(self, col_name: int | str, mode: str = 'w') -> Collection:
		"""
		:param mode: required permission level, either 'r' for read-only or 'rw' for read-write
		If a collection does not exist, it will be created and the proper mongoDB indexes will be set.
		"""

		if isinstance(col_name, int):
			col_name = str(col_name)

		if col_name in self.mongo_collections:
			if mode in self.mongo_collections[col_name]:
				return self.mongo_collections[col_name][mode]
		else:
			if col_name not in self.col_config:
				raise ValueError(f"Unknown collection: '{col_name}'")
			self.mongo_collections[col_name] = {}

		db_config = self._get_db_config(col_name)
		role = db_config.role.dict()[mode]
		db = self._get_pymongo_db(db_config.name, role=role)

		if 'w' in mode and col_name not in db.list_collection_names():
			try:
				self.create_ampel_collection(
					self.col_config[col_name], db_config.name, role
				)
			except CollectionInvalid:
				# raised when collection was created concurrently
				...
			except OperationFailure as exc:
				# also raised when collection was created concurrently
				if exc.code != 48: # NamespaceExists
					raise
		self.mongo_collections[col_name][mode] = db.get_collection(
			col_name,
			read_preference=ReadPreference.SECONDARY_PREFERRED if self.read_from_secondary else None
		)

		return self.mongo_collections[col_name][mode]


	def _get_pymongo_db(self, db_name: str, *, role: str) -> Database:
		"""
		:param db_name: without prefix
		"""

		if role not in self.mongo_clients:

			kwargs = self.mongo_options.dict()
			key = f'mongo/{role}'
			if self.vault and (ns := self.vault.get_named_secret(key, dict)):
				kwargs |= ns.get()

			try:
				self.mongo_clients[role] = MongoClient(self.mongo_uri, **kwargs)
			except ConfigurationError as exc:
				# hint at error source
				raise ConfigurationError(exc.args[0] + f' (from secret {key})', *exc.args[1:])

		return self.mongo_clients[role].get_database(
			f"{self.prefix}_{db_name}" if (db_name and not self.one_db) else self.prefix
		)


	def _get_db_config(self, col_name: str) -> AmpelDBModel:
		return next(
			filter(
				lambda x: self.col_config[col_name] in x.collections,
				self.databases
			)
		)


	def init_db(self) -> None:

		for db_config in self.databases:
			for col_config in db_config.collections:
				self.get_collection(col_config.name)

		self.get_collection('trace')
		self.get_collection('conf')


	def create_ampel_collection(self,
		col_config: AmpelColModel,
		db_name: str,
		role: str,
		logger: 'None | AmpelLogger' = None
	) -> Collection:
		"""
		:param resource_name: name of the AmpelConfig resource (resource.mongo) to be fed to MongoClient()
		:param db_name: name of the database to be used/created
		:param col_name: name of the collection to be created
		"""

		db = self._get_pymongo_db(db_name, role=role)

		if logger is None:
			# Avoid cyclic import error
			from ampel.log.AmpelLogger import AmpelLogger
			logger = AmpelLogger.get_logger()

		try:
			dbinfo = "[" + str(try_reduce(list(db.client.nodes))) + "]"
		except Exception:
			dbinfo = ""

		logger.info(f"Creating {db.name} -> {col_config.name} {dbinfo}")
		col = db.create_collection(col_config.name, **col_config.args)

		"""
		if col_config.name not in db.list_collection_names():
			# Create collection with custom args
			col = db.create_collection(col_config.name, **col_config.args)
		else:
			col = db.get_collection(col_config.name)
		"""

		if col_config.indexes:
			for idx in col_config.indexes:
				self._create_index(col, idx, logger)
	
		return col


	def set_col_index(self,
		col: Collection,
		config: AmpelColModel,
		logger: 'AmpelLogger',
		force_overwrite: bool = False
	) -> None:
		"""
		:param force_overwrite: delete index if it already exists.
		This can be useful if you want to change index options (for example: sparse=True/False)
		"""

		if not config.indexes:
			logger.info(f"No index data configured for collection {config.name}")
			return

		col_index_info = col.index_information()
		flat_indexes = []

		for idx in config.indexes:

			idx_id = idx.get_id()
			flat_indexes.append(idx_id)

			if idx_id in col_index_info:
				if force_overwrite:
					logger.info(f"  Deleting existing index: {idx_id}")
					col.drop_index(idx_id)
				else:
					logger.info(f"  Skipping already existing index: {idx_id}")
					continue

			self._create_index(col, idx, logger)

		for k in col_index_info:
			if k not in flat_indexes and k != "_id_":
				logger.info(f"  Removing index {k}")
				col.drop_index(k)


	def _create_index(self,
		col: Collection,
		index_data: IndexModel | ShortIndexModel,
		logger: 'AmpelLogger'
	) -> None:

		try:
			idx_params = index_data.dict(exclude_unset=True)
			logger.info(f"  Creating index for {col.database.name}.{col.name}: {idx_params}")
			col.create_index(idx_params['index'], **idx_params.get('args', {}))
		except Exception as e:
			logger.error(
				f"Index creation failed for '{col.database.name}.{col.name}': (args: {idx_params})",
				exc_info=e
			)


	def create_one_view(self,
		channel: ChannelId,
		logger: 'None | AmpelLogger' = None,
		force: bool = False
	) -> None:
		self.create_view(
			MongoOneView(channel=channel), str(channel),
			logger, force
		)


	def create_or_view(self,
		channels: Sequence[ChannelId],
		logger: 'None | AmpelLogger' = None,
		force: bool = False
	) -> None:

		if not isinstance(channels, collections.abc.Sequence) or len(channels) == 1:
			raise ValueError("Incorrect argument")

		self.create_view(
			MongoOrView(channel=channels),
			"_OR_".join(map(str, channels)),
			logger, force
		)


	def create_and_view(self,
		channels: Sequence[ChannelId],
		logger: 'None | AmpelLogger' = None,
		force: bool = False
	) -> None:

		if not isinstance(channels, collections.abc.Sequence) or len(channels) == 1:
			raise ValueError("Incorrect argument")

		self.create_view(
			MongoAndView(channel=channels),
			"_AND_".join(map(str, channels)),
			logger, force
		)


	def create_view(self,
		view: AbsMongoView,
		col_prefix: str,
		logger: 'None | AmpelLogger' = None,
		force: bool = False
	) -> None:

		db = self._get_pymongo_db("data", role="w")
		if force:
			col_names = db.list_collection_names()

		for el in ("stock", "t0", "t1", "t2", "t3"):

			agg = getattr(view, el)()

			if force and f'{col_prefix}_{el}' in col_names:
				if logger:
					logger.info(f"Discarding previous view {col_prefix}_{el}")
				db.drop_collection(f"{col_prefix}_{el}")

			if logger and logger.verbose > 1:
				from ampel.util.pretty import prettyjson
				logger.info(f"Collection {el} aggregation pipeline:")
				for line in prettyjson(agg).split("\n"):
					logger.info(line)
				logger.info("-"*80)

			db.create_collection(f'{col_prefix}_{el}', viewOn=el, pipeline=agg)


	def delete_one_view(self, channel: ChannelId, logger: 'None | AmpelLogger' = None) -> None:
		self.delete_view(str(channel), logger)


	def delete_or_view(self, channels: Sequence[ChannelId], logger: 'None | AmpelLogger' = None) -> None:
		if not isinstance(channels, collections.abc.Sequence) or len(channels) == 1:
			raise ValueError("Incorrect argument")
		self.delete_view("_OR_".join(map(str, channels)), logger)


	def delete_and_view(self, channels: Sequence[ChannelId], logger: 'None | AmpelLogger' = None) -> None:
		if not isinstance(channels, collections.abc.Sequence) or len(channels) == 1:
			raise ValueError("Incorrect argument")
		self.delete_view("_AND_".join(map(str, channels)), logger)


	def delete_view(self, view_prefix: str, logger: 'None | AmpelLogger' = None) -> None:

		db = self._get_pymongo_db("data", role="w")
		for el in ("stock", "t0", "t1", "t2", "t3"):
			db.drop_collection(f'{view_prefix}_{el}')


	def __repr__(self) -> str:
		return "<AmpelDB>"


	def drop_all_databases(self):
		for db in self.databases:
			pym_db = self._get_pymongo_db(db.name, role=db.role.w)
			pym_db.client.drop_database(pym_db.name)
		self.mongo_collections.clear()
		self.mongo_clients.clear()
		# deleting the attribute resets cached_property
		for attr in ("col_trace_ids", "col_conf_ids", "trace_ids", "conf_ids"):
			try:
				delattr(self, attr)
			except AttributeError:
				pass


	def add_trace_id(self, trace_id: int, arg: dict[str, Any]) -> None:

		# Save trace id to external collection
		if trace_id not in self.trace_ids:

			# Using try insert except on purpose because update_one/upsert does not maintain dict key order
			try:
				self.col_trace_ids.insert_one({'_id': trace_id} | arg)
			except DuplicateKeyError:
				pass

		self.trace_ids.add(trace_id)


	def add_conf_id(self, conf_id: int, arg: dict[str, Any]) -> None:

		# Save conf id to external collection
		if conf_id not in self.conf_ids:

			# Using try insert except on purpose because update_one/upsert does not maintain dict key order
			try:
				self.col_conf_ids.insert_one({'_id': conf_id} | arg)
			except DuplicateKeyError:
				pass

		self.conf_ids.add(conf_id)


def provision_accounts(ampel_db: AmpelDB, auth: dict[str, Any] = {}) -> dict[str, Any]:
	"""Create accounts required by the given Ampel configuration."""
	roles = defaultdict(list)
	for db in ampel_db.databases:
		name = f"{ampel_db.prefix}_{db.name}"
		roles[db.role.r].append({"db": name, "role": "read"})
		roles[db.role.w].append({"db": name, "role": "readWrite"})
	users = dict()
	admin: Database = MongoClient(ampel_db.mongo_uri, **auth).get_database("admin")
	tag = secrets.token_hex(8)
	for name, all_roles in roles.items():
		username = f"{name}-{tag}"
		password = secrets.token_hex()
		admin.command("createUser", username, pwd=password, roles=all_roles)
		users[f"mongo/{name}"] = {"username": username, "password": password}
	return users


def revoke_accounts(ampel_db: AmpelDB, auth: dict[str, Any] = {}) -> None:
	"""Delete accounts previously created with "provision"."""
	if ampel_db.vault is None:
		raise ValueError("No secrets vault configured")
	admin: Database = MongoClient(ampel_db.mongo_uri, **auth).get_database("admin")
	roles = {role for db in ampel_db.databases for role in db.role.dict().values()}
	for role in roles:
		if secret := ampel_db.vault.get_named_secret(f"mongo/{role}", dict):
			admin.command("dropUser", secret.get()["username"])
		else:
			raise ValueError(f"Unknown role '{role}'")


def list_accounts(ampel_db: AmpelDB, auth: dict[str, Any] = {}) -> dict[str, Any]:
	"""List configured accounts and roles."""
	return MongoClient(ampel_db.mongo_uri, **auth) \
		.get_database("admin") \
		.command("usersInfo")


def init_db(ampel_db: AmpelDB, auth: dict[str, Any] = {}) -> None:
	"""Initialize Ampel databases and collections."""
	if auth:
		print("="*40)
		print("DANGEROUS THINGS ARE ABOUT TO HAPPEN!")
		print("="*40)
		if input(f"Do you really want to reinitialize databases {[ampel_db.prefix+'_'+db.name for db in ampel_db.databases]} on {ampel_db.mongo_uri}? All data will be lost. Type 'yessir' to proceed: ") == "yessir":
			mc: MongoClient = MongoClient(ampel_db.mongo_uri, **auth)
			for db in ampel_db.databases:
				name = f"{ampel_db.prefix}_{db.name}"
				mc.drop_database(name)
			ampel_db.init_db()
		else:
			print("cancelled")
	else:
		ampel_db.init_db()


def main() -> None:
	import os
	import sys
	from argparse import ArgumentParser
	from getpass import getpass

	import yaml

	from ampel.core.AmpelContext import AmpelContext
	from ampel.secret.DictSecretProvider import DictSecretProvider

	parser = ArgumentParser(description="Manage access to Ampel databases")

	def from_env(key) -> None | str:
		if fname := os.environ.get(key + '_FILE', None):
			with open(fname, "r") as f:
				return f.read().rstrip()
		else:
			return os.environ.get(key, None)

	common = ArgumentParser(add_help=False)
	common.add_argument(
		'config_file_path',
		help="Path to an Ampel configuration file"
	)
	common.add_argument(
		'-u',
		'--username',
		type=str,
		default=from_env('MONGO_ROOT_USERNAME'),
		help="Mongo admin username",
	)
	common.add_argument(
		'-p',
		'--password',
		type=str,
		default=from_env('MONGO_ROOT_PASSWORD'),
		help="Mongo admin password",
	)

	subparsers = parser.add_subparsers(dest="command")
	subparsers.required = True

	def add_command(func, name=None):
		p = subparsers.add_parser(
			func.__name__ if not name else name,
			help=func.__doc__,
			parents=[common],
		)
		p.set_defaults(command=func)
		return p

	add_command(provision_accounts, "provision")

	p = add_command(revoke_accounts, "revoke")
	p.add_argument(
		'secrets',
		type=DictSecretProvider.load,
		default=None,
		help="Path to yaml file containing the accounts entries to be revoked",
	)

	add_command(list_accounts, "list")

	p = add_command(init_db, "init")
	p.add_argument(
		'--secrets',
		type=DictSecretProvider.load,
		default=None,
		help="Path to yaml file with auth secrets",
	)

	args = parser.parse_args()
	if args.username is not None:
		if args.password is None:
			args.password = getpass(prompt=f"Password for Mongo user {args.username}: ")
		auth = {
			"username": args.username,
			"password": args.password,
		}
	else:
		auth = {}

	ctx = AmpelContext.load(args.config_file_path, secrets=getattr(args, "secrets", None))

	if (ret := args.command(ctx.db, auth)):
		yaml.dump(ret, sys.stdout, sort_keys=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/MongoT0Ingester.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                23.05.2021
# Last Modified Date:  14.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from pymongo import UpdateOne
from typing import Any, Literal
from ampel.mongo.utils import maybe_use_each
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsDocIngester import AbsDocIngester


class MongoT0Ingester(AbsDocIngester[DataPoint]):
	""" Inserts `DataPoint` into the t0 collection  """

	#: If 1 or 2: raise DuplicateKeyError on attempts to upsert the same id with different bodies
	#: 1: partial check is performed (compatible with potential muxer projection)
	#: 2: strict check is performed
	extended_match: Literal[0, 1, 2] = 0

	def ingest(self, doc: DataPoint) -> None:

		match: dict[str, Any] = {'id': doc['id']}
		upd: dict[str, Any] = {
			'$addToSet': {
				'channel': maybe_use_each(doc['channel'])
			},
			'$push': {
				'meta': maybe_use_each(doc['meta']) # meta must be set by compiler
			}
		}

		if 'body' in doc:
			upd['$setOnInsert'] = {'body': doc['body']}
			if self.extended_match:
				if self.extended_match == 2:
					match['body'] = doc['body']
				else:
					for k, v in doc['body'].items():
						match['body.' + k] = v

		if 'origin' in doc:
			match['origin'] = doc['origin']

		if 'tag' in doc:
			upd['$addToSet']['tag'] = maybe_use_each(doc['tag'])

		if 'stock' in doc:
			upd['$addToSet']['stock'] = doc['stock'] if isinstance(doc['stock'], (int, bytes, str)) \
				else maybe_use_each(doc['stock'])

		self.updates_buffer.add_t0_update(
			UpdateOne(match, upd, upsert=True)
		)



import asyncio

import pytest

from ampel.core.AmpelController import AmpelController


@pytest.mark.asyncio
async def test_run_cancel(dev_context):
    c = AmpelController(dev_context.config, tier=2)
    r = asyncio.create_task(c.run())
    await asyncio.sleep(0.1)
    r.cancel()
    await r




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/query/t2.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.02.2018
# Last Modified Date:  27.12.2019
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from bson.binary import Binary
from typing import Any
from ampel.types import StockId, ChannelId, StrictIterable, strict_iterable
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.mongo.utils import maybe_match_array
from ampel.mongo.query.general import build_general_query


def build_stateless_query(
	stock: None | StockId | StrictIterable[StockId] = None,
	channel: None | ChannelId | dict | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId] = None,
	t2_subsel: None | int | str | StrictIterable[int | str] = None
) -> dict[str, Any]:
	"""
	| Builds a pymongo query dict aiming at loading transient t2 or compounds docs \
	| Stateless query: all avail compounds and t2docs (although possibly \
	constrained by parameter t2_subsel) are targeted.

	:param stock: (query can be performed on multiple ids at once)

	:param channel: see :obj:`QueryMatchSchema <ampel.query.QueryMatchSchema>` for details. \
	None (no criterium) means all channels are considered.

	:param t2_subsel: optional sub-selection of t2 records based on t2 class names. \
	-> only t2 records matching with the provided t2 class names will be returned. \
	If None (or empty iterable): all t2 docs associated with the matched transients will be targeted. \
	"""

	query = build_general_query(stock=stock, channel=channel)

	if t2_subsel:
		query['unit'] = t2_subsel if isinstance(t2_subsel, (str, int)) \
			else maybe_match_array(t2_subsel)

	return query


def build_statebound_t1_query(
	states: str | bytes | Binary | StrictIterable[str | bytes | Binary],
) -> dict[str, Any]:
	return {'_id': get_compound_match(states)}


def build_statebound_t2_query(
	stock: None | StockId | StrictIterable[StockId],
	states: str | bytes | Binary | StrictIterable[str | bytes | Binary],
	channel: None | ChannelId | dict | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId] = None,
	t2_subsel: None | int | str | StrictIterable[int | str] = None
) -> dict[str, Any]:
	"""
	See :func:`build_stateless_query <build_stateless_query>` docstring
	"""

	query = build_general_query(stock=stock, channel=channel)
	query['link'] = get_compound_match(states)

	if t2_subsel:
		query['unit'] = t2_subsel if isinstance(t2_subsel, (str, int)) \
			else maybe_match_array(t2_subsel)

	return query

def _to_binary(st: str | bytes | Binary) -> Binary:
	if isinstance(st, str):
		if len(st) == 32:
			return Binary(bytes.fromhex(st), 0)
		else:
			raise ValueError("Provided state strings must have 32 characters")
	elif isinstance(st, bytes):
		if len(st) == 16:
			return Binary(st, 0)
		else:
			raise ValueError("Provided state bytes must have a length of 16")
	else:
		if st.subtype == 0:
			return st
		else:
			raise ValueError("Bson Binary states must have subtype 0")


def get_compound_match(
	states: str | bytes | Binary | StrictIterable[str | bytes | Binary],
) -> Binary | dict[str, list[Binary]]:
	"""
	:raises ValueError: if provided states parameter is invalid
	"""

	# Single state was provided
	if isinstance(states, (str, bytes, Binary)):
		return _to_binary(states)

	# Multiple states were provided
	elif isinstance(states, strict_iterable):

		return {
			'$in': [_to_binary(st) for st in states]
		}

	else:
		raise ValueError(
			f"Type of provided state ({type(states)}) must be "
			f"bytes, str, bson.Binary or sequences of these"
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/query/var/LogsMatcher.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.11.2018
# Last Modified Date:  20.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import collections.abc
from typing import Any, get_args
from collections.abc import Sequence
from datetime import datetime
from bson.objectid import ObjectId
from ampel.types import ChannelId, StockId
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.util.collections import check_seq_inner_type
from ampel.mongo.schema import apply_schema
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf


class LogsMatcher:


	@classmethod
	def new(cls,
		after: None | str = None,
		before: None | str = None,
		channel: None | ChannelId | Sequence[ChannelId] = None,
		stock: None | StockId | Sequence[StockId] = None,
		run: None | int | Sequence[int] = None,
		custom: None | dict[str, Any] = None,
		id_mapper: None | AbsIdMapper = None,
		flag: None | int = None,
		**kwargs # ignored, added for cli convenience
	) -> "LogsMatcher":

		matcher = cls()

		if after:
			matcher.set_after(after)

		if before:
			matcher.set_before(before)

		if channel:
			matcher.set_channel(channel)

		if id_mapper:
			matcher.set_id_mapper(id_mapper)

		if stock:
			matcher.set_stock(stock)

		if run:
			matcher.set_run(run)

		if custom:
			matcher.set_custom(
				next(iter(custom.keys())),
				next(iter(custom.values()))
			)

		if flag:
			matcher.set_flag(flag)

		return matcher


	def __init__(self) -> None:
		self.mcrit: dict[str, Any] = {}
		self.id_mapper: None | AbsIdMapper = None


	def get_match_criteria(self) -> dict[str, Any]:
		return self.mcrit


	def set_id_mapper(self, id_mapper: AbsIdMapper) -> None:
		self.id_mapper = id_mapper


	def set_flag(self, arg) -> 'LogsMatcher':
		""" Mongo's operator "$bitsAllSet" is used """
		self.mcrit['f'] = {"$bitsAllSet": arg}
		return self


	def set_channel(self,
		channel: ChannelId | Sequence[ChannelId] | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId],
		compact_logs: bool = True
	) -> 'LogsMatcher':


		if isinstance(channel, get_args(ChannelId)):
			v: Any = channel
		elif isinstance(channel, (dict, AllOf, AnyOf, OneOf)):
			v = apply_schema(self.mcrit, 'channel', channel)
		elif isinstance(channel, collections.abc.Sequence):
			v = {'$in': channel}
		else:
			raise ValueError(f"Unsupported type for parameter channel: {type(channel)}")


		if compact_logs:
			self.mcrit['$or'] = [{'c': v}, {'m.c': v}]
		else:
			self.mcrit['c'] = v

		return self


	def set_stock(self,
		stock_id: StockId | Sequence[StockId]
	) -> 'LogsMatcher':
		if (
			self.id_mapper and (
				isinstance(stock_id, str) or
				check_seq_inner_type(stock_id, str)
			)
		):
			self.mcrit['s'] = self.id_mapper.to_ampel_id(stock_id) \
				if isinstance(stock_id, str) else {'$in': self.id_mapper.to_ampel_id(stock_id)} # type: ignore
		else:
			self.mcrit['s'] = stock_id if isinstance(stock_id, get_args(StockId)) else {'$in': stock_id}
		return self


	def set_run(self, run_id: int | Sequence[int] | dict) -> 'LogsMatcher':
		if isinstance(run_id, dict):
			self.mcrit['r'] = run_id # enables -run-json '{"$gt": 1}'
		else:
			self.mcrit['r'] = run_id if isinstance(run_id, int) else {'$in': run_id}
		return self


	def set_custom(self, key: str, value: Any) -> 'LogsMatcher':
		if (isinstance(value, collections.abc.Sequence) and not isinstance(value, str)):
			self.mcrit[key] = {'$in': value}
		else:
			self.mcrit[key] = value
		return self


	def set_after(self, dt: datetime | str) -> 'LogsMatcher':
		"""
		Note: time operation is greater than / *equals*
		:param dt: either datetime object or string (datetime.fromisoformat is used)
		"""
		return self._set_time_constraint(dt, '$gte')


	def set_before(self, dt: datetime | str) -> 'LogsMatcher':
		"""
		Note: time operation is before than / *equals*
		:param dt: either datetime object or string (datetime.fromisoformat is used)
		"""
		return self._set_time_constraint(dt, '$lte')


	def _set_time_constraint(self, dt: datetime | str, op: str) -> 'LogsMatcher':
		"""
		Note: time operation is greater than / *equals*
		:param dt: either datetime object or string (datetime.fromisoformat is used)
		"""

		if isinstance(dt, datetime):
			pass
		elif isinstance(dt, str):
			dt = datetime.fromisoformat(dt)
		else:
			raise ValueError()

		if "_id" not in self.mcrit:
			self.mcrit["_id"] = {}

		self.mcrit["_id"][op] = ObjectId.from_datetime(dt)

		return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/ProcessConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  02.01.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class ProcessConfigCollector(AbsDictConfigCollector):

	def add(self,
		arg: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:

		# Doing basic validation here already
		for k in ("name", "schedule"):
			if k not in arg:
				return self.missing_key(
					what="Process", key=k, dist_name=dist_name, register_file=register_file
				)

		proc_name = arg["name"]

		if 'tier' not in arg:
			if len(self.tier) == 2 and self.tier.startswith('t') and self.tier[1].isdigit():
				arg['tier'] = int(self.tier[1])
			else:
				arg['tier'] = None

		if self.verbose:
			self.logger.log(
				VERBOSE, f"Adding {self.tier} process: '{proc_name}' from {register_file}"
			)

		if self.check_duplicates(
			proc_name, dist_name, version, register_file,
			section_detail = f"{arg['tier']}.process"
		):
			return

		self.__setitem__(proc_name, arg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/T3SimpleStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.04.2021
# Last Modified Date:  17.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator

from ampel.types import OneOrMany, Annotated
from ampel.struct.T3Store import T3Store
from ampel.model.UnitModel import UnitModel
from ampel.content.T3Document import T3Document
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.t3.stage.T3ThreadedStager import T3ThreadedStager
from ampel.t3.stage.SimpleViewGenerator import SimpleViewGenerator


class T3SimpleStager(T3ThreadedStager):
	"""
	To be used in combination with MongoViews.
	Otherwise, use T3ProjectingStager.
	"""

	#: t3 units (AbsT3ReviewUnit) to execute
	execute: OneOrMany[Annotated[UnitModel, AbsT3ReviewUnit]]

	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self.units: list[AbsT3ReviewUnit] = []

		if self.logger.verbose > 1:
			self.logger.debug("Setting up T3SimpleStager")

		for exec_def in [self.execute] if isinstance(self.execute, UnitModel) else self.execute:
			self.units.append(self.get_unit(exec_def))


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> None | Generator[T3Document, None, None]:

		if len(self.units) == 1:
			return self.proceed(
				self.units[0],
				SimpleViewGenerator(self.units[0], gen, self.stock_updr, self.context.config),
				t3s
			)

		return self.proceed_threaded(self.units, gen, t3s)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/demo/DemoFirstPointT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2020
# Last Modified Date:  15.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit


class DemoFirstPointT2Unit(AbsPointT2Unit):

	ingest = {'filter': 'PPSFilter', 'sort': 'jd', 'select': 'first'}
	chatty: bool = False

	def process(self, datapoint: DataPoint) -> UBson | UnitResult:
		if self.chatty:
			self.logger.info(f"Parsing {datapoint['id']}")
		return {"time": time()}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/AliasConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  02.01.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class AliasConfigCollector(AbsDictConfigCollector):
	"""
	Aliases are "tier" scoped and by default scoped
	at the (package) distribution level as well
	"""

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.global_alias: dict[str, Any] = {}


	def add(self,
		arg: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:

		if not isinstance(arg, dict):
			self.error(
				f"{self.tier} alias value must be a dict. "
				f"Offending value {arg}\n"
				f"{self.distrib_hint(dist_name, register_file)}"
			)
			return

		for k, v in arg.items():

			if "/" in k:
				self.error(
					f"{self.tier} alias cannot contain '/'."
					f"Offending key {k}\n"
					f"{self.distrib_hint(dist_name, register_file)}"
				)
				continue

			try:

				key = k

				# Global alias
				if k and k[0] == "%":
					scope = "global"
					self.global_alias[key] = dist_name
				else:
					# Distribution scoped alias
					key = f"{dist_name}/{k}"
					scope = "scoped"

				if self.verbose:
					self.logger.log(VERBOSE,
						f"Adding {scope} {self.tier} alias: {key}"
					)

				if self.check_duplicates(
					key, dist_name, version, register_file,
					section_detail = f"{self.tier} {scope} alias"
				):
					continue

				self.__setitem__(key, v)

			except Exception as e:
				self.error(
					f"Error occured while loading {self.tier} alias {key} " +
					self.distrib_hint(dist_name, register_file),
					exc_info=e
				)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/ingest/ChainedIngestionHandler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.05.2020
# Last Modified Date:  24.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from typing import Literal, Any
from collections.abc import Callable, Sequence
from ampel.types import StockId, ChannelId, UnitId, DataPointId, UBson, Tag
from ampel.abstract.AbsT0Muxer import AbsT0Muxer
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsApplicable import AbsApplicable
from ampel.abstract.AbsT0Unit import AbsT0Unit
from ampel.abstract.AbsT1ComputeUnit import AbsT1ComputeUnit
from ampel.abstract.AbsT1CombineUnit import AbsT1CombineUnit
from ampel.abstract.AbsT1RetroCombineUnit import AbsT1RetroCombineUnit
from ampel.model.UnitModel import UnitModel
from ampel.enum.DocumentCode import DocumentCode
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.enum.JournalActionCode import JournalActionCode
from ampel.model.ingest.CompilerOptions import CompilerOptions
from ampel.model.ingest.T1Combine import T1Combine
from ampel.model.ingest.T1CombineCompute import T1CombineCompute
from ampel.model.ingest.T1CombineComputeNow import T1CombineComputeNow
from ampel.model.ingest.T2Compute import T2Compute
from ampel.model.ingest.IngestDirective import IngestDirective
from ampel.model.ingest.DualIngestDirective import DualIngestDirective
from ampel.model.ingest.IngestBody import IngestBody
from ampel.model.DPSelection import DPSelection
from ampel.core.AmpelContext import AmpelContext
from ampel.base.LogicalUnit import LogicalUnit
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.content.StockDocument import StockDocument
from ampel.content.DataPoint import DataPoint
from ampel.content.MetaActivity import MetaActivity
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer
from ampel.ingest.T0Compiler import T0Compiler
from ampel.ingest.T2Compiler import T2Compiler
from ampel.ingest.StockCompiler import StockCompiler
from ampel.ingest.T1Compiler import T1Compiler
from ampel.struct.UnitResult import UnitResult
from ampel.struct.T1CombineResult import T1CombineResult
from ampel.log import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.log.handlers.DefaultRecordBufferingHandler import DefaultRecordBufferingHandler
from ampel.util.hash import build_unsafe_dict_id


class T2Block:
	__slots__ = "unit", "config", "slc", "sort", "filter", "group"
	unit: UnitId
	config: None | int
	filter: None | AbsApplicable
	sort: None | Callable
	slc: None | slice
	group: None | Sequence[int]


class T1ComputeBlock:
	__slots__ = 'unit', 'unit_name', 'config', 'trace_id'
	unit: None | AbsT1ComputeUnit
	unit_name: None | UnitId
	config: None | int
	trace_id: None | int


class T1CombineBlock:
	__slots__ = 'unit', 'trace_id', 'compute', 'channel', 'group', 'state_t2', 'point_t2'
	unit: AbsT1CombineUnit | AbsT1RetroCombineUnit
	trace_id: None | int
	compute: T1ComputeBlock
	channel: ChannelId
	group: None | Sequence[int]
	state_t2: None | list[T2Block]
	point_t2: None | list[T2Block]


class T0MuxBlock:
	__slots__ = 'unit', 'config', 'trace_id', 'combine', 'point_t2'
	unit: AbsT0Muxer
	config: None | int
	trace_id: None | int
	combine: None | list[T1CombineBlock]  # mux.combine
	point_t2: None | list[T2Block] # mux.insert.point_t2


class IngestBlock:
	__slots__ = 'channel', 'mux', 'combine', 'combine', 'point_t2', 'stock_t2'
	channel: ChannelId
	mux: None | T0MuxBlock
	combine: None | list[T1CombineBlock] # combine blocks
	point_t2: None | list[T2Block] # point t2
	stock_t2: None | list[T2Block] # stock t2


T1CombineCache = dict[
	tuple[AbsT1CombineUnit | AbsT1RetroCombineUnit, tuple[DataPointId, ...]],
	tuple[list[list[DataPointId] | T1CombineResult], set[ChannelId]]
]

T1ComputeCache = dict[
	tuple[AbsT1ComputeUnit, tuple[DataPointId, ...]],
	tuple[UBson | UnitResult, StockId]
]


class ChainedIngestionHandler:

	__slots__ = '__dict__', 'updates_buffer', 'ingest_stats', 'iblocks', 'shaper', \
		'shaper_trace_id', 't0_compiler', 't1_compiler', 'stock_compiler', \
		'state_t2_compiler', 'point_t2_compiler', 'stock_t2_compiler', \
		'stock_ingester', 't0_ingester', 't1_ingester', 't2_ingester'

	def __init__(self,
		context: AmpelContext,
		shaper: UnitModel,
		directives: Sequence[IngestDirective | DualIngestDirective],
		updates_buffer: DBUpdatesBuffer,
		run_id: int,
		trace_id: dict[str, None | int],
		tier: Literal[-1, 0, 1, 2, 3],
		compiler_opts: CompilerOptions,
		logger: AmpelLogger,
		database: str = "mongo",
		origin: None | int = None,
		int_time: bool = True,
		include_extra_meta: int = 2
	):
		"""
		:param trace_id: base trace if of root caller (such as AlertConsumer or T1Generator)
		:param int_time: timestamp accuracy for journal/meta (otherwise float)
		:param include_extra_meta: example for extra: {'alert': 231273612673162}.
		- 0: do not include provided extra meta data into the generated meta entries
		- 1: include provided extra meta data in all generated meta entries of all docs except point t2 docs
		- 2: include provided extra meta data in all generated meta entries of all docs
		Note: granularity might increase in the future
		"""

		self.updates_buffer = updates_buffer
		self.logger = logger
		self.context = context
		self.run_id = run_id
		self.int_time = int_time
		self.include_extra_meta = include_extra_meta
		self.base_trace_id = trace_id

		self.iblocks: list[tuple[IngestBlock, IngestBlock]] = []
		self.ingest_stats: list[float] = []
		self._mux_cache: dict[int, AbsT0Muxer] = {}
		self._t1_combine_units_cache: dict[int, AbsT1CombineUnit | AbsT1RetroCombineUnit] = {}
		self._t1_compute_units_cache: dict[int, AbsT1ComputeUnit] = {}

		if not directives:
			raise ValueError("Need at least 1 directive")

		self.shaper = self.context.loader.new_logical_unit(
			model=shaper, logger=logger, sub_type=AbsT0Unit
		)
		self.shaper_trace_id = self.shaper._trace_id

		# Base compiler parameters
		bopts: dict[str, Any] = {"origin": origin, "tier": tier, "run_id": run_id}

		# Create compilers
		self.t0_compiler = T0Compiler(**(compiler_opts.t0 | bopts))
		self.t1_compiler = T1Compiler(**(compiler_opts.t1 | bopts))
		self.stock_compiler = StockCompiler(**(compiler_opts.stock | bopts))
		self.state_t2_compiler = T2Compiler(**(compiler_opts.state_t2 | bopts))
		self.point_t2_compiler = T2Compiler(**(compiler_opts.point_t2 | bopts), col="t0")
		self.stock_t2_compiler = T2Compiler(**(compiler_opts.stock_t2 | bopts), col='stock')

		# Create ingesters
		dbconf = self.context.config.get(f'{database}.ingest', dict, raise_exc=True)
		def get_ingester_model(key: str) -> UnitModel:
			model = dbconf[key]
			if isinstance(model, str):
				return UnitModel(unit=model)
			else:
				return UnitModel(**model)
		self.t0_ingester = AuxUnitRegister.new_unit(
			model = get_ingester_model('t0'),
			sub_type = AbsDocIngester[DataPoint],
			updates_buffer = updates_buffer
		)

		self.t1_ingester = AuxUnitRegister.new_unit(
			model = get_ingester_model('t1'),
			sub_type = AbsDocIngester[T1Document],
			updates_buffer = updates_buffer
		)

		self.t2_ingester = AuxUnitRegister.new_unit(
			model = get_ingester_model('t2'),
			sub_type = AbsDocIngester[T2Document],
			updates_buffer = updates_buffer
		)

		self.stock_ingester = AuxUnitRegister.new_unit(
			model = get_ingester_model('stock'),
			sub_type = AbsDocIngester[StockDocument],
			updates_buffer = updates_buffer
		)

		for directive in directives:
			self.iblocks.append(
				self._new_ingest_blocks(directive, updates_buffer, logger)
			)


	def _new_ingest_blocks(self,
		directive: IngestDirective | DualIngestDirective,
		updates_buffer: DBUpdatesBuffer,
		logger: AmpelLogger
	) -> tuple[IngestBlock, IngestBlock]:

		if isinstance(directive, DualIngestDirective):
			known_ib = self._new_ingest_block(
				directive.ingest['known'], directive.channel, updates_buffer, logger
			)
			new_ib = self._new_ingest_block(
				directive.ingest['new'], directive.channel, updates_buffer, logger
			)
			return known_ib, new_ib

		elif isinstance(directive, IngestDirective):
			ib = self._new_ingest_block(
				directive.ingest, directive.channel, updates_buffer, logger
			)
			return ib, ib

		else:
			ValueError("Unknown directive type")
			

	def _new_ingest_block(self,
		directive: IngestBody,
		channel: ChannelId,
		updates_buffer: DBUpdatesBuffer,
		logger: AmpelLogger
	) -> IngestBlock:

		ib = IngestBlock()
		ib.mux = None
		ib.combine = None
		ib.stock_t2 = None
		ib.point_t2 = None
		ib.channel = channel

		if directive.mux:

			muxb = ib.mux = T0MuxBlock()
			muxb.trace_id = 0
			muxb.combine = None
			muxb.point_t2 = None

			i = build_unsafe_dict_id(directive.mux.dict(exclude_unset=False))
			if i in self._mux_cache:
				muxer = self._mux_cache[i]
			else:

				# We create a logger associated a buffering handler
				# whose logs entries are later transfered to the main logger
				buf_hdlr = DefaultRecordBufferingHandler(level=self.logger.level)
				buf_logger = AmpelLogger.get_logger(
					base_flag = (getattr(self.logger, 'base_flag', 0) & ~LogFlag.CORE) | LogFlag.UNIT,
					console = False,
					handlers = [buf_hdlr]
				)

				# Spawn new instance
				self._mux_cache[i] = muxer = self.context.loader.new_context_unit(
					model = directive.mux, context = self.context, sub_type = AbsT0Muxer,
					logger = buf_logger, updates_buffer = updates_buffer
				)

				# Shortcut to avoid muxer.logger.handlers[?]
				setattr(muxer, '_buf_hdlr', buf_hdlr)

			muxb.unit = muxer
			muxb.trace_id = muxer._trace_id

			if directive.mux.combine:

				# States, and T2s based thereon
				muxb.combine = [
					self._setup_t1_combine(channel, t1_combine)
					for t1_combine in directive.mux.combine
				]

			if directive.mux.insert:
				ib.mux.point_t2 = [
					self._gen_t2_block(el)
					for el in directive.mux.insert['point_t2']
				]

		if directive.combine:

			# States, and T2s based thereon
			ib.combine = [
				self._setup_t1_combine(channel, t1_combine)
				for t1_combine in directive.combine
			]

		# (Data)Point T2s
		if directive.point_t2:
			ib.point_t2 = [self._gen_t2_block(el) for el in directive.point_t2]

		# Stock T2s
		if directive.stock_t2:
			ib.stock_t2 = [self._gen_t2_block(el) for el in directive.stock_t2]

		return ib


	def _setup_t1_combine(self,
		channel: ChannelId,
		t1_combine: T1Combine | T1CombineCompute | T1CombineComputeNow
	) -> T1CombineBlock:
		"""
		Add the ingesters specified in ``t1_combine`` to ``cache``, reusing
		existing instances if possible.
		
		:param t1_combine: subclause of ingestion directive
		:param channel: channel of parent directive
		"""

		t1b = T1CombineBlock()
		t1b.trace_id = 0
		t1b.channel = channel
		t1b.group = None
		t1b.state_t2 = None
		t1b.point_t2 = None

		# Avoid 'ifs' later on if initiliazed with Nones
		t1b.compute = T1ComputeBlock()
		t1b.compute.unit = None
		t1b.compute.unit_name = None
		t1b.compute.config = None

		# Cache t1 combine units
		i = build_unsafe_dict_id(t1_combine.dict(exclude_unset=False))
		if i in self._t1_combine_units_cache:
			t1_unit = self._t1_combine_units_cache[i]
		else:

			# We create a logger associated a buffering handler
			# whose logs entries are later transfered to the main logger
			buf_hdlr = DefaultRecordBufferingHandler(level=self.logger.level)
			buf_logger = AmpelLogger.get_logger(
				base_flag = (getattr(self.logger, 'base_flag', 0) & ~LogFlag.CORE) | LogFlag.UNIT,
				console = False,
				handlers = [buf_hdlr]
			)

			chan_dict = self.context.config.get(f'channel.{channel}', dict, raise_exc=True)
			t1_unit = self.context.loader.new_logical_unit(
				model = t1_combine, # model.config can potentially include/define channel (for chan-specific exclusions)
				logger = buf_logger,
				sub_type = AbsT1CombineUnit,
				access = chan_dict['access'],
				policy = chan_dict['policy']
			)

			# Shortcut to avoid t1_unit.logger.handlers[?]
			setattr(t1_unit, '_buf_hdlr', buf_hdlr)
			self._t1_combine_units_cache[t1b.trace_id] = t1_unit

		t1b.unit = t1_unit
		t1b.trace_id = t1_unit._trace_id
		if g := t1_combine.group:
			t1b.group = g if isinstance(g, Sequence) else [g]

		# State T2s (are defined along with t1 directives usually)
		# We allow the definition of multiple combiners t2 ingesters
		if isinstance(t1_combine, (T1Combine, T1CombineComputeNow)):
			if t1_combine.state_t2:
				t1b.state_t2 = [self._gen_t2_block(el) for el in t1_combine.state_t2]
			if t1_combine.point_t2:
				t1b.point_t2 = [self._gen_t2_block(el) for el in t1_combine.point_t2]

		if isinstance(t1_combine, (T1CombineCompute, T1CombineComputeNow)):

			t1b.compute = T1ComputeBlock()
			t1b.compute.unit_name = t1_combine.compute.unit
			if isinstance(t1_combine.compute.config, int):
				t1b.compute.config = t1_combine.compute.config
			else:
				raise ValueError("Integer expected for t1_combine.compute.config")

			# On the fly t1 computation requested
			if isinstance(t1_combine, T1CombineComputeNow):

				# Cache t1 compute units
				i = build_unsafe_dict_id(t1_combine.compute.dict(exclude_unset=False))
				if i in self._t1_compute_units_cache:
					t1_compute_unit = self._t1_compute_units_cache[i]
				else:
					t1_compute_unit = self.context.loader.new_logical_unit(
						model = t1_combine,
						logger = buf_logger,
						sub_type = AbsT1ComputeUnit
					)
					self._t1_compute_units_cache[i] = t1_compute_unit

				t1b.compute.unit = t1_compute_unit
				t1b.compute.trace_id = t1_compute_unit._trace_id

		return t1b


	def _gen_t2_block(self, im: T2Compute) -> T2Block:

		ingest_opts: dict[str, Any] = {}
		if not (t2_info := self.context.config.get(f'unit.{im.unit}', dict)):
			raise ValueError(f'Unknown T2 unit {im.unit}')

		# The unit is installed locally
		if 'fqn' in t2_info:
			ingest_opts = getattr(
				self.context.loader.get_class_by_name(name=im.unit, unit_type=LogicalUnit),
				'eligible', DPSelection()
			).dict()

		# Ingest options build up (dict.update operation is used):
		# 1) Static class member 'eligible' (for example, T2CatMatch might define:
		#    eligible: ClassVar[DPSelection] = DPSelection(filter="PPSFilter", sort="jd", select="first")
		# 2) Specific unit configuration 'ingest' (field defined in T2Compute) might define:
		#    {'ingest': {"filter": None}}
		# In which case the first datapoint of the list sorted base of field 'jd' will be selected
		# Note: that an explicit None is required as ingest options are not hard overridden (dict.update)

		if im.ingest:

			if isinstance(im.ingest, str):
				if im.ingest not in self.context.config._config['alias']['t2']:
					raise ValueError(f"Ingest alias {im.ingest} not found")
				ingest_opts.update(
					self.context.config._config['alias']['t2'][im.ingest]
				)
			else: # AmpelBaseModel
				ingest_opts.update(im.ingest.dict())

		ib = T2Block()
		ib.unit = im.unit # type: ignore[assignment]
		ib.config = im.config # type: ignore[assignment]

		# Save confid to external collection for posterity
		if isinstance(im.config, int):
			# Make sure config is valid and do provenance check
			self.context.loader.get_init_config(im.config)

		# Only for point t2 units (which can customize the ingestion)
		if ingest_opts:
			ib.filter, ib.sort, ib.slc = DPSelection(**ingest_opts).tools()
		else:
			ib.filter = ib.sort = ib.slc = None

		if im.group:
			ib.group = [im.group] if isinstance(im.group, int) else im.group
		else:
			ib.group = None

		return ib


	def ingest(self,
		alert_dps: Sequence[dict[str, Any]],
		filter_results: list[tuple[int, bool | int]],
		stock_id = 0,
		tag: None | Tag | list[Tag] = None,
		jm_extra: None | dict[str, Any] = None,
		stock_body: None | dict[str, Any] = None
	) -> None:
		"""
		Create database documents.
		:param filter_results: the value returned from
		  :func:`~ampel.abstract.AbsAlertFilter.AbsAlertFilter.process` if alert was accepted
		:param jm_extra: extra info to be added to journal or meta enties. Ex: {'alert_id': 123}
		"""

		with self.updates_buffer.group_updates():
			ingest_start = time()

			# process *modifies* dict instances loaded by fastavro
			dps: list[DataPoint] = self.shaper.process(alert_dps, stock_id)

			if not dps: # Not sure if this can happen
				return

			add_other_tag: None | MetaActivity = {'action': MetaActionCode.ADD_OTHER_TAG, 'tag': tag} if tag else None

			# Set of chans (last parameter) is used for logging
			mux_cache: dict[AbsT0Muxer, tuple[None | list[DataPoint], None | list[DataPoint], set[ChannelId]]] = {}
			t1_comb_cache: T1CombineCache = {}
			t1_comp_cache: T1ComputeCache = {}

			# ingestion blocks
			ibs = self.iblocks

			for i, fres in filter_results:

				# Add alert and shaper version info to stock journal entry
				jentry: dict[str, Any] = {
					'action': JournalActionCode.STOCK_ADD_CHANNEL | JournalActionCode.STOCK_BUMP_UPD,
					'traceid': self.base_trace_id | {'shaper': self.shaper_trace_id}
				}

				if jm_extra:
					jentry = jm_extra | jentry

				if i > 0: # Known stock (for the current channel)
					ib = ibs[i][0]
				else: # New stock
					ib = ibs[-i][1]

				# Muxer requested
				if mux := ib.mux:

					# Add muxer version info to stock journal entry
					jentry['traceid']['muxer'] = mux.trace_id

					# Potentially load previous results from cache
					if mux.unit in mux_cache:
						dps_insert, dps_combine, s = mux_cache[mux.unit]
						s.add(ib.channel)
					else:
						dps_insert, dps_combine = mux.unit.process(dps, stock_id)
						mux_cache[mux.unit] = dps_insert, dps_combine, {ib.channel}

					if dps_combine:

						if x := [
							dp for dp in dps_combine
							if 'channel' in dp and ib.channel not in dp['channel']
						]:
							dps_insert = (dps_insert + x) if dps_insert else x

					if dps_insert:

						self.t0_compiler.add(dps_insert, ib.channel, self.shaper_trace_id, jm_extra)

						# TODO: make this addition optional (a stock with a million dps would create pblms)
						jentry['upsert'] = [el['id'] for el in dps_insert]
						jentry['action'] |= JournalActionCode.T0_ADD_CHANNEL

						if mux.point_t2:
							jentry['action'] |= JournalActionCode.T2_ADD_CHANNEL
							self.ingest_point_t2s(
								dps_insert, fres, stock_id, ib.channel, mux.point_t2, add_other_tag,
								jm_extra if self.include_extra_meta > 1 else None
							)

					# Muxed T1 and associated T2 ingestions
					if dps_combine and mux.combine:
						self.ingest_t12(
							dps_combine, fres, stock_id, jentry, mux.combine,
							t1_comb_cache, t1_comp_cache, add_other_tag,
							jm_extra if self.include_extra_meta else None
						)

				else:
					self.t0_compiler.add(dps, ib.channel, self.shaper_trace_id)

				# Non-muxed T1 and associated T2 ingestions
				if ib.combine:
					self.ingest_t12(
						dps, fres, stock_id, jentry, ib.combine, t1_comb_cache, t1_comp_cache,
						add_other_tag, meta_extra = jm_extra if self.include_extra_meta else None
					)
					
				# Non-muxed point T2s
				if ib.point_t2:
					self.ingest_point_t2s(
						dps, fres, stock_id, ib.channel, ib.point_t2, add_other_tag,
						jm_extra if self.include_extra_meta > 1 else None
					)

				# Stock T2s
				if ib.stock_t2:
					for t2b in ib.stock_t2:
						self.stock_t2_compiler.add(
							t2b.unit, t2b.config, stock_id, stock_id,
							ib.channel, self.base_trace_id, add_other_tag,
							jm_extra if self.include_extra_meta else None
						)

				# Flush potential unit logs
				###########################

				logger = self.logger
				for muxer, (_, _, chans) in mux_cache.items():
					if muxer._buf_hdlr.buffer: # type: ignore[attr-defined]
						muxer._buf_hdlr.forward( # type: ignore[attr-defined]
							logger, stock=stock_id, channel=list(chans), extra = jm_extra
						)

				for (t1_unit, _), (_, chans) in t1_comb_cache.items():
					if t1_unit._buf_hdlr.buffer: # type: ignore[union-attr]
						t1_unit._buf_hdlr.forward( # type: ignore[union-attr]
							logger, stock=stock_id, channel=list(chans), extra = jm_extra
						)

				if not self.stock_compiler.register:
					self.stock_compiler.add(
						stock_id, ib.channel, journal = jentry, # type: ignore[arg-type]
						tag = add_other_tag['tag'] if add_other_tag else None # type: ignore[arg-type]
					)

			# Commit
			########
			now = int(time()) if self.int_time else time()

			self.t0_compiler.commit(self.t0_ingester, now)

			if self.t1_compiler.t1s:
				self.t1_compiler.commit(self.t1_ingester, now)

			if self.stock_t2_compiler.t2s:
				self.stock_t2_compiler.commit(self.t2_ingester, now)

			if self.point_t2_compiler.t2s:
				self.point_t2_compiler.commit(self.t2_ingester, now)

			if self.state_t2_compiler.t2s:
				self.state_t2_compiler.commit(self.t2_ingester, now)

			self.stock_compiler.commit(self.stock_ingester, now, body=stock_body)
			self.ingest_stats.append(time() - ingest_start)


	def ingest_point_t2s(self,
		dps: list[DataPoint],
		fres: bool | int,
		stock_id: StockId,
		channel: ChannelId,
		state_t2: list[T2Block],
		add_other_tag: None | MetaActivity = None,
		meta_extra: None | dict[str, Any] = None
	) -> None:

		for t2b in state_t2:

			# Filter group selection / veto
			if t2b.group and isinstance(fres, int) and fres not in t2b.group:
				continue

			# filter (ex: use only photopoints or upperlimis)
			f = t2b.filter.apply(dps) if t2b.filter else dps

			# Sort (ex: by body.jd)
			if t2b.sort:
				f = t2b.sort(f)

			# Slice (ex: first datapoint)
			if t2b.slc:
				f = f[t2b.slc]

			if isinstance(f, list):
				for el in f:
					self.point_t2_compiler.add(
						t2b.unit, t2b.config, stock_id, el['id'], channel,
						self.base_trace_id, add_other_tag, meta_extra
					)
			else:
				self.point_t2_compiler.add(
					t2b.unit, t2b.config, stock_id, f['id'], channel,
					self.base_trace_id, add_other_tag, meta_extra
				)


	def ingest_t12(self,
		dps: list[DataPoint], fres: bool | int, stock_id: StockId,
		jentry: dict[str, Any], t1bs: list[T1CombineBlock],
		t1_comb_cache: T1CombineCache, t1_comp_cache: T1ComputeCache,
		add_other_tag: None | MetaActivity = None,
		meta_extra: None | dict[str, Any] = None,
	) -> None:

		tdps = tuple(el['id'] for el in dps)

		# Loop through t1 blocks
		for t1b in t1bs:

			# Skip unmatched group
			if t1b.group and isinstance(fres, int) and fres not in t1b.group:
				continue

			# Potentially load previous results from cache
			if (t1b.unit, tdps) in t1_comb_cache:
				lres, s = t1_comb_cache[(t1b.unit, tdps)]
				s.add(t1b.channel)
			else:
				comb_res = t1b.unit.combine(iter(dps))
				if isinstance(comb_res, T1CombineResult): # case T1CombineResult
					lres = [comb_res]
				elif isinstance(comb_res, list):
					if len(comb_res) == 0:
						lres = []
					elif isinstance(comb_res[0], DataPointId): # case list[DataPointId]
						lres = [comb_res] # type: ignore[list-item]
					else:
						# case list[list[DataPointId]], list[T1CombineResult]
						lres = comb_res # type: ignore[assignment]
				t1_comb_cache[(t1b.unit, tdps)] = lres, {t1b.channel}

			# T1 combine(...) can return multiple subsets of the initial datapoints
			for tres in lres:

				body = None
				tid: dict[str, Any] = self.base_trace_id | {'combiner': t1b.trace_id}

				if 'muxer' in jentry['traceid']:
					tid['muxer'] = jentry['traceid']['muxer']

				mx = meta_extra.copy() if meta_extra else {}
				macts: list[MetaActivity] = [
					{
						'action': MetaActionCode.ADD_CHANNEL | MetaActionCode.BUMP_STOCK_UPD,
						'channel': t1b.channel
					}
				]

				if isinstance(tres, T1CombineResult):
					t1_dps = tres.dps
					if tres.meta:
						mx |= tres.meta
						jentry['action'] |= JournalActionCode.T1_EXTRA_META
						macts[0]['action'] |= MetaActionCode.EXTRA_META
					if tres.code:
						code = tres.code
						jentry['action'] |= JournalActionCode.T1_SET_CODE
						macts[0]['action'] |= MetaActionCode.SET_UNIT_CODE
						if t1b.compute.unit:
							macts[0]['code'] = code
					else:
						code = DocumentCode.OK
						macts[0]['action'] |= MetaActionCode.SET_CODE
				else:
					t1_dps = tres
					code = DocumentCode.OK
					macts[0]['action'] |= MetaActionCode.SET_CODE

				if excl := [el['id'] for el in dps if el['id'] not in t1_dps]:
					macts[0]['action'] |= MetaActionCode.ADD_T1_EXCL
					macts[0]['excl'] = excl

				je = jentry.copy()
				je['traceid'] = jentry['traceid'].copy()
				je['traceid']['combiner'] = t1b.trace_id

				if not t1_dps:
					self.logger.info(f"No datapoints returned by t1 unit ({t1b.channel})")
					if stock_id:
						je['combine_empty'] = True
						self.stock_compiler.add(
							stock_id, t1b.channel, journal = je, # type: ignore[arg-type]
							tag = add_other_tag['tag'] if add_other_tag else None # type: ignore[arg-type]
						)
					continue

				# On the fly t1 computation requested
				if t1b.compute.unit:

					je['traceid']['t1_compute'] = t1b.compute.trace_id
					macts.append({'action': MetaActionCode.ADD_BODY})

					# Potentially load previous results from "t1 compute" cache
					k = t1b.compute.unit, tuple(t1_dps)
					if k in t1_comp_cache:
						t1_res = t1_comp_cache[k]
					else:
						t1_res = t1_comp_cache[k] = t1b.compute.unit.compute(
							[dp for dp in dps if dp['id'] in t1_dps]
						)

					# AbsT1ComputeUnit can be used to determine stock
					stock_id = t1_res[1]

					if isinstance(t1_res[0], UnitResult):
						body = t1_res[0].body
						if t1_res[0].journal:
							je |= t1_res[0].journal.dict()
							je['action'] |= JournalActionCode.T1_EXTRA_JOURNAL
							macts[1]['action'] |= MetaActionCode.EXTRA_JOURNAL
						if t1_res[0].code:
							code = t1_res[0].code
							je['action'] |= JournalActionCode.T1_SET_CODE
							macts[1]['action'] |= MetaActionCode.SET_UNIT_CODE
						else:
							code = DocumentCode.NEW
							macts[1]['action'] |= MetaActionCode.SET_CODE
					else:
						body = t1_res[0]
						code = DocumentCode.NEW
						macts[1]['action'] |= MetaActionCode.SET_CODE

				je['action'] |= JournalActionCode.T1_ADD_CHANNEL
				mx['code'] = code

				# Note: we ignore potential stock from T1 result here
				link = self.t1_compiler.add(
					t1_dps,
					t1b.channel,
					tid,
					stock_id,
					meta_extra = mx,
					activity = macts,
					unit = t1b.compute.unit_name,
					config = t1b.compute.config,
					body = body,
					code = code
				)

				je['link'] = link

				self.stock_compiler.add(
					stock_id, t1b.channel, journal = je, # type: ignore[arg-type]
					tag = add_other_tag['tag'] if add_other_tag else None # type: ignore[arg-type]
				)

				if t1b.state_t2:

					jentry['action'] |= JournalActionCode.T2_ADD_CHANNEL
					for t2b in t1b.state_t2:

						# Skip unmatched group
						if t2b.group and isinstance(fres, int) and fres not in t2b.group:
							continue

						self.state_t2_compiler.add(
							t2b.unit,
							t2b.config,
							stock_id,
							link,
							t1b.channel,
							tid,
							add_other_tag,
							meta_extra if self.include_extra_meta else None
						)

				if t1b.point_t2:
				
					jentry['action'] |= JournalActionCode.T2_ADD_CHANNEL
					self.ingest_point_t2s(
						[el for el in dps if el['id'] in t1_dps],
						fres, stock_id, t1b.channel, t1b.point_t2, add_other_tag,
						meta_extra if self.include_extra_meta > 1 else None
					)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/db/MongoClientOptionsModel.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                08.11.2020
# Last Modified Date:  08.11.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

from ampel.base.AmpelBaseModel import AmpelBaseModel

class MongoClientOptionsModel(AmpelBaseModel):

	# 0 means use operating system's default socket timeout
	socketTimeoutMS: int = 0

	# 0 means use operating system's default socket timeout
	connectTimeoutMS: int = 0

	# https://github.com/mongodb/specifications/blob/master/source/server-selection/server-selection.rst#serverselectiontimeoutms
	serverSelectionTimeoutMS: int = 30000 # default is 30,000 (milliseconds)

	# Connect directly to target host, rather than the primary of its replica
	# set. This is useful when connecting e.g. through an ssh tunnel to hosts
	# that are not externally resolvable
	directConnection: bool = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/unit/T3LogAggregatedStocks.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.05.2022
# Last Modified Date:  17.07.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.struct.T3Store import T3Store
from ampel.abstract.AbsT3PlainUnit import AbsT3PlainUnit


class T3LogAggregatedStocks(AbsT3PlainUnit):

	input_unit: str = "T3AggregatingStager"

	def process(self, t3s: None | T3Store = None):

		self.logger.info(f"Running {self.__class__.__name__}")

		if not t3s:
			raise ValueError("A T3 store is required, please check your config")

		self.logger.info(
			"Stock ids",
			extra = {
				'stock': [
					int(k) for k in t3s \
						.get_mandatory_view(unit=self.input_unit) \
						.get_body(raise_exc=True)
				]
			}
		)

		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/db/IndexModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                19.10.2019
# Last Modified Date:  13.04.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.mongo.model.FieldModel import FieldModel

class IndexModel(AmpelBaseModel):

	index: Sequence[FieldModel]
	args: None | dict

	def get_id(self) -> str:
		"""
		Returns an indexId similar to what pymongo index_information outputs.
		Ex: [('tranId', 1), ('channel', 1)] -> tranId_1_channel_1
		"""
		return "_".join(el.get_id() for el in self.index)
